SENTINEL2_FILENAME_REGEX = r"""
    ^(?P<tile>T\d{2}[A-Z]{3})
    _(?P<date>\d{8})T\d{6}
    _(?P<band>B[018][\dA]|SCL)
    (?:_(?P<resolution>\d+)m)?
    .*
    \..*$
"""


SENTINEL2_IMAGE_REGEX = r"""
    ^(?P<mission_id>S2[AB])
    _MSI(?P<level>[A-Z]\d{1}[A-Z])
    _(?P<date>\d{8})T\d{6}
    _(?P<baseline>N\d{4})
    _(?P<orbit>R\d{3})
    _(?P<tile>T\d{2}[A-Z]{3})
    _\d{8}T\d{6}
    .*
    .*$
"""

SENTINEL2_MOSAIC_FILENAME_REGEX = r"""
    ^SENTINEL2X_
    (?P<date>\d{8})
    .*?
    (?P<tile>T\d{2}[A-Z]{3})
    .*?
    _(?P<band>B\d{1,2}A?)_
    .*?
    (?:_(?P<resolution>\d+m))?
    .*?\.tif$
"""

SENTINEL2_MOSAIC_IMAGE_REGEX = r"""
    ^SENTINEL2X_
    (?P<date>\d{8})
    -\d{6}
    -\d{3}
    _(?P<level>[A-Z]\d{1}[A-Z])
    .*(?P<tile>T\d{2}[A-Z]{3})
    .*.*$
"""


SENTINEL2_SCL_CLASSES = {
    0: "No Data (Missing data)",  # 000000
    1: "Saturated or defective pixel",  # ff0000
    2: "Topographic casted shadows",  # 2f2f2f
    3: "Cloud shadows",  # 643200
    4: "Vegetation",  # 00a000
    5: "Not-vegetated",  # ffe65a
    6: "Water",  # 0000ff
    7: "Unclassified",  # 808080
    8: "Cloud medium probability",  # c0c0c0
    9: "Cloud high probability",  # ffffff
    10: "Thin cirrus",  # 64c8ff
    11: "Snow or ice",  # ff96ff
}



import functools
from abc import ABC
from abc import abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

import joblib
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from shapely import Geometry
from shapely import STRtree
from shapely import get_parts
from shapely import make_valid
from shapely import union_all
from shapely.errors import GEOSException

from .utils import _unary_union_for_notna
from .utils import make_valid_and_keep_geom_type


@dataclass
class AbstractRunner(ABC):
    """Blueprint for 'runner' classes.

    Subclasses must implement a 'run' method.

    Args:
        n_jobs: Number of workers.
        backend: Backend for the workers.
    """

    n_jobs: int
    backend: str | None = None

    @abstractmethod
    def run(self, *args, **kwargs) -> Any:
        """Abstract run method."""

    def __str__(self) -> str:
        """String representation."""
        return (
            f"{self.__class__.__name__}(n_jobs={self.n_jobs}, backend='{self.backend}')"
        )


@dataclass
class UnionRunner(AbstractRunner):
    """Run shapely.union_all with pandas.groupby.

    Subclasses must implement a 'run' method that takes the arguments
    'df' (GeoDataFrame or GeoSeries), 'by' (optional column to group by), 'grid_size'
    (passed to shapely.union_all) and **kwargs passed to pandas.DataFrame.groupby.
    Defaults to None, meaning the default runner with number of workers set
    to 'n_jobs'.


    Args:
        n_jobs: Number of workers.
        backend: Backend for the workers.
    """

    n_jobs: int
    backend: str | None = None

    def run(
        self,
        df: GeoDataFrame | GeoSeries | pd.DataFrame | pd.Series,
        by: str | list[str] | None = None,
        grid_size: float | int | None = None,
        **kwargs,
    ) -> GeoSeries | GeoDataFrame:
        """Run groupby on geometries in parallel (if n_jobs > 1)."""
        # assume geometry column is 'geometry' if input is pandas.Series og pandas.DataFrame
        try:
            geom_col: str = df.geometry.name
        except AttributeError:
            try:
                geom_col: str | None = df.name
                if geom_col is None:
                    geom_col = "geometry"
            except AttributeError:
                geom_col = "geometry"
        try:
            crs = df.crs
        except AttributeError:
            crs = None

        unary_union_for_grid_size = functools.partial(
            _unary_union_for_notna, grid_size=grid_size
        )

        as_index = kwargs.pop("as_index", True)
        if by is None and "level" not in kwargs:
            by = np.zeros(len(df), dtype="int64")

        try:
            # (Geo)DataFrame
            groupby_obj = df.groupby(by, **kwargs)[geom_col]
        except KeyError:
            # (Geo)Series
            groupby_obj = df.groupby(by, **kwargs)

        if self.n_jobs is None or self.n_jobs == 1:
            results = groupby_obj.agg(unary_union_for_grid_size)
            index = results.index
        else:
            backend = self.backend or "loky"
            with joblib.Parallel(n_jobs=self.n_jobs, backend=backend) as parallel:
                results = parallel(
                    joblib.delayed(unary_union_for_grid_size)(geoms)
                    for _, geoms in groupby_obj
                )
            index = groupby_obj.size().index
        agged = GeoSeries(results, index=index, name=geom_col, crs=crs)
        if not as_index:
            return agged.reset_index()
        return agged


@dataclass
class GridSizeUnionRunner(UnionRunner):
    """Run shapely.union_all with pandas.groupby for different grid sizes until no GEOSException is raised.

    Subclasses must implement a 'run' method that takes the arguments
    'df' (GeoDataFrame or GeoSeries), 'by' (optional column to group by), 'grid_size'
    (passed to shapely.union_all) and **kwargs passed to pandas.DataFrame.groupby.
    Defaults to None, meaning the default runner with number of workers set
    to 'n_jobs'.


    Args:
        n_jobs: Number of workers.
        backend: Backend for the workers.
    """

    n_jobs: int
    backend: str | None = None
    grid_sizes: list[float | int] | None = None

    def __post_init__(self) -> None:
        """Check that grid_sizes is passed."""
        if self.grid_sizes is None:
            raise ValueError(
                f"must set 'grid_sizes' in the {self.__class__.__name__} initialiser."
            )

    def run(
        self,
        df: GeoDataFrame | GeoSeries | pd.DataFrame | pd.Series,
        by: str | list[str] | None = None,
        grid_size: int | float | None = None,
        **kwargs,
    ) -> GeoSeries | GeoDataFrame:
        """Run groupby on geometries in parallel (if n_jobs > 1) with grid_sizes."""
        try:
            return super().run(df, by=by, grid_size=grid_size, **kwargs)
        except GEOSException:
            pass
        for i, grid_size in enumerate(self.grid_sizes):
            try:
                return super().run(df, by=by, grid_size=grid_size, **kwargs)
            except GEOSException as e:
                if i == len(self.grid_sizes) - 1:
                    raise e


def _strtree_query(
    arr1: np.ndarray,
    arr2: np.ndarray,
    method: str,
    indices1: np.ndarray | None = None,
    indices2: np.ndarray | None = None,
    **kwargs,
):
    tree = STRtree(arr2)
    func = getattr(tree, method)
    left, right = func(arr1, **kwargs)
    if indices1 is not None:
        index_mapper1 = {i: x for i, x in enumerate(indices1)}
        left = np.array([index_mapper1[i] for i in left])
    if indices2 is not None:
        index_mapper2 = {i: x for i, x in enumerate(indices2)}
        right = np.array([index_mapper2[i] for i in right])
    return left, right


@dataclass
class RTreeQueryRunner(AbstractRunner):
    """Run shapely.STRTree chunkwise.

    Subclasses must implement a 'query' method that takes a numpy.ndarray
    of geometries as 0th and 1st argument and **kwargs passed to the query method,
    chiefly 'predicate' and 'distance'. The 'query' method should return a tuple
    of two arrays representing the spatial index pairs of the left and right input arrays.
    Defaults to None, meaning the default runner with number of workers set
    to 'n_jobs'.

    Args:
        n_jobs: Number of workers.
        backend: Backend for the workers.
    """

    n_jobs: int
    backend: str = "loky"

    def run(
        self, arr1: np.ndarray, arr2: np.ndarray, method: str = "query", **kwargs
    ) -> tuple[np.ndarray, np.ndarray]:
        """Run a spatial rtree query and return indices of hits from arr1 and arr2 in a tuple of two arrays."""
        if (
            (self.n_jobs or 1) > 1
            and len(arr1) / self.n_jobs > 10_000
            and len(arr1) / len(arr2)
        ):
            chunks = np.array_split(np.arange(len(arr1)), self.n_jobs)
            assert sum(len(x) for x in chunks) == len(arr1)
            with joblib.Parallel(self.n_jobs, backend=self.backend) as parallel:
                results = parallel(
                    joblib.delayed(_strtree_query)(
                        arr1[chunk],
                        arr2,
                        method=method,
                        indices1=chunk,
                        **kwargs,
                    )
                    for chunk in chunks
                )
            left = np.concatenate([x[0] for x in results])
            right = np.concatenate([x[1] for x in results])
            return left, right
        elif (
            (self.n_jobs or 1) > 1
            and len(arr2) / self.n_jobs > 10_000
            and len(arr2) / len(arr1)
        ):
            chunks = np.array_split(np.arange(len(arr2)), self.n_jobs)
            with joblib.Parallel(self.n_jobs, backend=self.backend) as parallel:
                results = parallel(
                    joblib.delayed(_strtree_query)(
                        arr1,
                        arr2[chunk],
                        method=method,
                        indices2=chunk,
                        **kwargs,
                    )
                    for chunk in chunks
                )
            left = np.concatenate([x[0] for x in results])
            right = np.concatenate([x[1] for x in results])
            return left, right

        return _strtree_query(arr1, arr2, method=method, **kwargs)


@dataclass
class OverlayRunner(AbstractRunner):
    """Run a vectorized shapely overlay operation on two equal-length numpy arrays.

    Subclasses must implement a 'run' method that takes an overlay function (shapely.intersection, shapely.difference etc.)
    as 0th argument and two numpy.ndarrays of same length as 1st and 2nd argument.
    The 'run' method should also take the argument 'grid_size' to be passed to the overlay function
    and the argument 'geom_type' which is used to keep only relevant geometries (polygon, line or point)
    in cases of GEOSExceptions caused by geometry type mismatch.
    Defaults to an instance of OverlayRunner, which is run sequencially (no n_jobs)
    because the vectorized shapely functions are usually faster than any attempt to parallelize.
    """

    n_jobs: None = None
    backend: None = None

    def run(
        self,
        func: Callable,
        arr1: np.ndarray,
        arr2: np.ndarray,
        grid_size: int | float | None,
        geom_type: str | None,
    ) -> np.ndarray:
        """Run the overlay operation (func) with fallback.

        First tries to run func, then, if GEOSException, geometries are made valid
        and only geometries with correct geom_type (point, line, polygon) are kept
        in GeometryCollections.
        """
        try:
            return func(arr1, arr2, grid_size=grid_size)
        except GEOSException:
            arr1 = make_valid_and_keep_geom_type(arr1, geom_type=geom_type)
            arr2 = make_valid_and_keep_geom_type(arr2, geom_type=geom_type)
            arr1 = arr1.loc[lambda x: x.index.isin(arr2.index)].to_numpy()
            arr2 = arr2.loc[lambda x: x.index.isin(arr1.index)].to_numpy()
            return func(arr1, arr2, grid_size=grid_size)


@dataclass
class GridSizeOverlayRunner(OverlayRunner):
    """Run a shapely overlay operation rowwise for different grid_sizes until success."""

    n_jobs: int
    backend: str | None
    grid_sizes: list[float | int] | None = None

    def __post_init__(self) -> None:
        """Check that grid_sizes is passed."""
        if self.grid_sizes is None:
            raise ValueError(
                f"must set 'grid_sizes' in the {self.__class__.__name__} initialiser."
            )

    def run(
        self,
        func: Callable,
        arr1: np.ndarray,
        arr2: np.ndarray,
        grid_size: int | float | None = None,
        geom_type: str | None = None,
    ) -> np.ndarray:
        """Run the overlay operation rowwise with fallback.

        The overlay operation (func) is looped for each row in arr1 and arr2
        as 0th and 1st argument to 'func' and 'grid_size' as keyword argument. If a GEOSException is thrown,
        geometries are made valid and GeometryCollections are forced to either
        (Multi)Point, (Multi)Polygon or (Multi)LineString, depending on the value in "geom_type".
        Then, if Another GEOSException is thrown, the overlay operation is looped for the grid_sizes given
        in the instance's 'grid_sizes' attribute.

        """
        kwargs = dict(
            grid_size=grid_size,
            geom_type=geom_type.lower() if geom_type is not None else None,
            grid_sizes=self.grid_sizes,
        )
        with joblib.Parallel(self.n_jobs, backend="threading") as parallel:
            return parallel(
                joblib.delayed(_run_overlay_rowwise)(func, g1, g2, **kwargs)
                for g1, g2 in zip(arr1, arr2, strict=True)
            )


def _fix_gemetry_fast(geom: Geometry, geom_type: str | None) -> Geometry:
    geom = make_valid(geom)
    if geom.geom_type == geom_type or geom_type is None:
        return geom
    return union_all([g for g in get_parts(geom) if geom_type in g.geom_type])


def _run_overlay_rowwise(
    func: Callable,
    geom1: Geometry,
    geom2: Geometry,
    grid_size: float | int | None,
    geom_type: str | None,
    grid_sizes: list[float | int],
) -> Geometry:
    try:
        return func(geom1, geom2, grid_size=grid_size)
    except GEOSException:
        pass
    geom1 = _fix_gemetry_fast(geom1, geom_type)
    geom2 = _fix_gemetry_fast(geom2, geom_type)
    try:
        return func(geom1, geom2)
    except GEOSException:
        pass
    for i, grid_size in enumerate(grid_sizes):
        try:
            return func(geom1, geom2, grid_size=grid_size)
        except GEOSException as e:
            if i == len(grid_sizes) - 1:
                raise e



import functools
import itertools
import numbers
import warnings
from collections.abc import Hashable
from collections.abc import Iterable
from typing import Any

import numpy as np
import pandas as pd
import pyproj
import shapely
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from geopandas.array import GeometryArray
from geopandas.array import GeometryDtype
from numpy.typing import NDArray
from shapely import Geometry
from shapely import extract_unique_points
from shapely import get_coordinates
from shapely import get_parts
from shapely import linestrings
from shapely import make_valid
from shapely.errors import GEOSException
from shapely.geometry import LineString
from shapely.geometry import MultiPoint
from shapely.geometry import Point

from .conversion import coordinate_array
from .conversion import to_bbox
from .conversion import to_gdf
from .conversion import to_geoseries
from .geometry_types import get_geom_type
from .geometry_types import make_all_singlepart
from .geometry_types import to_single_geom_type
from .neighbors import get_k_nearest_neighbors
from .sfilter import sfilter
from .sfilter import sfilter_split


def split_geom_types(gdf: GeoDataFrame | GeoSeries) -> tuple[GeoDataFrame | GeoSeries]:
    return tuple(
        gdf[gdf.geom_type == geom_type] for geom_type in gdf.geom_type.unique()
    )


def get_common_crs(
    iterable: Iterable[Hashable], strict: bool = False
) -> pyproj.CRS | None:
    """Returns the common not-None crs or raises a ValueError if more than one.

    Args:
        iterable: Iterable of objects with the attribute "crs" or a list
            of CRS-like (pyproj.CRS-accepted) objects.
        strict: If False (default), falsy CRS-es will be ignored and None
            will be returned if all CRS-es are falsy. If strict is True,

    Returns:
        pyproj.CRS object or None (if all crs are None).

    Raises:
        ValueError if there are more than one crs. If strict is True,
        None is included.
    """
    crs = set()
    for obj in iterable:
        try:
            crs.add(obj.crs)
        except AttributeError:
            pass

    if not crs:
        try:
            crs = list(set(iterable))
        except TypeError:
            return None

    truthy_crs = list({x for x in crs if x})

    if strict and len(truthy_crs) != len(crs):
        raise ValueError("Mix of falsy and truthy CRS-es found.")

    if len(truthy_crs) > 1:
        # sometimes the bbox is slightly different, resulting in different
        # hash values for same crs. Therefore, trying to
        actually_different = set()
        for x in truthy_crs:
            x = pyproj.CRS(x)
            if x.to_string() in {j.to_string() for j in actually_different}:
                continue
            actually_different.add(x)

        if len(actually_different) == 1:
            return next(iter(actually_different))
        raise ValueError("'crs' mismatch.", truthy_crs)

    return pyproj.CRS(truthy_crs[0])


def is_bbox_like(obj: Any) -> bool:
    if (
        hasattr(obj, "__iter__")
        and len(obj) == 4
        and all(isinstance(x, numbers.Number) for x in obj)
    ):
        return True
    return False


def is_wkt(text: str) -> bool:
    gemetry_types = ["point", "polygon", "line", "geometrycollection"]
    return any(x in text.lower() for x in gemetry_types)


def _push_geom_col(gdf: GeoDataFrame) -> GeoDataFrame:
    """Makes the geometry column the rightmost column in the GeoDataFrame.

    Args:
        gdf: GeoDataFrame.

    Returns:
        The GeoDataFrame with the geometry column pushed all the way to the right.
    """
    geom_col = gdf._geometry_column_name
    return gdf.reindex(columns=[c for c in gdf.columns if c != geom_col] + [geom_col])


def drop_inactive_geometry_columns(gdf: GeoDataFrame) -> GeoDataFrame:
    """Removes geometry columns in a GeoDataFrame if they are not active."""
    for col in gdf.columns:
        if (
            isinstance(gdf[col].dtype, GeometryDtype)
            and col != gdf._geometry_column_name
        ):
            gdf = gdf.drop(col, axis=1)
    return gdf


def _rename_geometry_if(gdf: GeoDataFrame) -> GeoDataFrame:
    geom_col = gdf._geometry_column_name
    if geom_col == "geometry" and geom_col in gdf.columns:
        return gdf
    elif geom_col in gdf.columns:
        return gdf.rename_geometry("geometry")

    geom_cols = list(
        {col for col in gdf.columns if isinstance(gdf[col].dtype, GeometryDtype)}
    )
    if len(geom_cols) == 1:
        gdf._geometry_column_name = geom_cols[0]
        return gdf.rename_geometry("geometry")

    raise ValueError(
        "There are multiple geometry columns and none are the active geometry"
    )


def clean_geoms(
    gdf: GeoDataFrame | GeoSeries,
    ignore_index: bool = False,
) -> GeoDataFrame | GeoSeries:
    """Fixes geometries, then removes empty, NaN and None geometries.

    Args:
        gdf: GeoDataFrame or GeoSeries to be cleaned.
        ignore_index: If True, the resulting axis will be labeled 0, 1, …, n - 1.
            Defaults to False

    Returns:
        GeoDataFrame or GeoSeries with fixed geometries and only the rows with valid,
        non-empty and not-NaN/-None geometries.

    Examples:
    ---------
    >>> import sgis as sg
    >>> import pandas as pd
    >>> from shapely import wkt
    >>> gdf = sg.to_gdf([
    ...         "POINT (0 0)",
    ...         "LINESTRING (1 1, 2 2)",
    ...         "POLYGON ((3 3, 4 4, 3 4, 3 3))"
    ...         ])
    >>> gdf
                                                geometry
    0                            POINT (0.00000 0.00000)
    1      LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....

    Add None and empty geometries.

    >>> missing = pd.DataFrame({"geometry": [None]})
    >>> empty = sg.to_gdf(wkt.loads("POINT (0 0)").buffer(0))
    >>> gdf = pd.concat([gdf, missing, empty])
    >>> gdf
                                                geometry
    0                            POINT (0.00000 0.00000)
    1      LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....
    0                                               None
    0                                      POLYGON EMPTY

    Clean.

    >>> sg.clean_geoms(gdf)
                                                geometry
    0                            POINT (0.00000 0.00000)
    1      LINESTRING (1.00000 1.00000, 2.00000 2.00000)
    2  POLYGON ((3.00000 3.00000, 4.00000 4.00000, 3....
    """
    warnings.filterwarnings("ignore", "GeoSeries.notna", UserWarning)

    if isinstance(gdf, GeoDataFrame):
        # only repair if necessary
        if not gdf.geometry.is_valid.all():
            try:
                gdf.geometry = gdf.make_valid()
            except GEOSException as e:
                raise type(e)(f"{e}: {gdf.geometry}") from e

        notna = gdf.geometry.notna()
        if not notna.all():
            gdf = gdf.loc[notna]

        is_empty = gdf.geometry.is_empty
        if is_empty.any():
            gdf = gdf.loc[~is_empty]

    elif isinstance(gdf, GeoSeries):
        if not gdf.is_valid.all():
            gdf = gdf.make_valid()

        notna = gdf.notna()
        if not notna.all():
            gdf = gdf.loc[notna]

        is_empty = gdf.is_empty
        if is_empty.any():
            gdf = gdf.loc[~is_empty]

    else:
        raise TypeError(f"'gdf' should be GeoDataFrame or GeoSeries, got {type(gdf)}")

    if ignore_index:
        gdf = gdf.reset_index(drop=True)

    return gdf


def get_grouped_centroids(
    gdf: GeoDataFrame, groupby: str | list[str], as_string: bool = True
) -> pd.Series:
    """Get the centerpoint of the geometries within a group.

    Args:
        gdf: GeoDataFrame.
        groupby: column to group by.
        as_string: If True (default), coordinates are returned in
            the format "{x}_{y}". If False, coordinates are returned
            as Points.

    Returns:
        A pandas.Series of grouped centroids with the index of 'gdf'.
    """
    centerpoints = gdf.assign(geometry=lambda x: x.centroid)

    grouped_centerpoints = centerpoints.dissolve(by=groupby).assign(
        geometry=lambda x: x.centroid
    )
    xs = grouped_centerpoints.geometry.x
    ys = grouped_centerpoints.geometry.y

    if as_string:
        grouped_centerpoints["wkt"] = [
            f"{int(x)}_{int(y)}" for x, y in zip(xs, ys, strict=False)
        ]
    else:
        grouped_centerpoints["wkt"] = [
            Point(x, y) for x, y in zip(xs, ys, strict=False)
        ]

    return gdf[groupby].map(grouped_centerpoints["wkt"])


def sort_large_first(gdf: GeoDataFrame | GeoSeries) -> GeoDataFrame | GeoSeries:
    """Sort GeoDataFrame by area in decending order.

    Args:
        gdf: A GeoDataFrame or GeoSeries.

    Returns:
        A GeoDataFrame or GeoSeries sorted from large to small in area.

    Examples:
    ---------
    Create GeoDataFrame with NaN values.

    >>> import sgis as sg
    >>> df = sg.to_gdf(
    ...     [
    ...         (0, 1),
    ...         (1, 0),
    ...         (1, 1),
    ...         (0, 0),
    ...         (0.5, 0.5),
    ...     ]
    ... )
    >>> df.geometry = df.buffer([4, 1, 2, 3, 5])
    >>> df["col"] = [None, 1, 2, None, 1]
    >>> df["col2"] = [None, 1, 2, 3, None]
    >>> df["area"] = df.area
    >>> df
                                                geometry  col  col2       area
    0  POLYGON ((4.56136 0.53436, 4.54210 0.14229, 4....  NaN   NaN  50.184776
    1  POLYGON ((1.40111 0.71798, 1.39630 0.61996, 1....  1.0   1.0   3.136548
    2  POLYGON ((2.33302 0.49287, 2.32339 0.29683, 2....  2.0   2.0  12.546194
    3  POLYGON ((3.68381 0.46299, 3.66936 0.16894, 3....  NaN   3.0  28.228936
    4  POLYGON ((5.63590 0.16005, 5.61182 -0.33004, 5...  1.0   NaN  78.413712

    >>> sg.sort_large_first(df)
                                                geometry  col  col2       area
    4  POLYGON ((5.63590 0.16005, 5.61182 -0.33004, 5...  1.0   NaN  78.413712
    0  POLYGON ((4.56136 0.53436, 4.54210 0.14229, 4....  NaN   NaN  50.184776
    3  POLYGON ((3.68381 0.46299, 3.66936 0.16894, 3....  NaN   3.0  28.228936
    2  POLYGON ((2.33302 0.49287, 2.32339 0.29683, 2....  2.0   2.0  12.546194
    1  POLYGON ((1.40111 0.71798, 1.39630 0.61996, 1....  1.0   1.0   3.136548

    >>> sg.sort_nans_last(sg.sort_large_first(df))
                                                geometry  col  col2       area
    2  POLYGON ((2.33302 0.49287, 2.32339 0.29683, 2....  2.0   2.0  12.546194
    1  POLYGON ((1.40111 0.71798, 1.39630 0.61996, 1....  1.0   1.0   3.136548
    4  POLYGON ((5.63590 0.16005, 5.61182 -0.33004, 5...  1.0   NaN  78.413712
    3  POLYGON ((3.68381 0.46299, 3.66936 0.16894, 3....  NaN   3.0  28.228936
    0  POLYGON ((4.56136 0.53436, 4.54210 0.14229, 4....  NaN   NaN  50.184776
    """
    # using enumerate, then iloc on the sorted dict keys.
    # to avoid creating a temporary area column (which doesn't work for GeoSeries).
    area_mapper = dict(enumerate(gdf.area.values))
    sorted_areas = dict(reversed(sorted(area_mapper.items(), key=_get_dict_value)))
    return gdf.iloc[list(sorted_areas)]


def sort_long_first(gdf: GeoDataFrame | GeoSeries) -> GeoDataFrame | GeoSeries:
    """Sort GeoDataFrame by length in decending order.

    Args:
        gdf: A GeoDataFrame or GeoSeries.

    Returns:
        A GeoDataFrame or GeoSeries sorted from long to short in length.
    """
    # using enumerate, then iloc on the sorted dict keys.
    # to avoid creating a temporary area column (which doesn't work for GeoSeries).
    length_mapper = dict(enumerate(gdf.length.values))
    sorted_lengths = dict(reversed(sorted(length_mapper.items(), key=_get_dict_value)))
    return gdf.iloc[list(sorted_lengths)]


def sort_short_first(gdf: GeoDataFrame | GeoSeries) -> GeoDataFrame | GeoSeries:
    """Sort GeoDataFrame by length in ascending order.

    Args:
        gdf: A GeoDataFrame or GeoSeries.

    Returns:
        A GeoDataFrame or GeoSeries sorted from short to long in length.
    """
    # using enumerate, then iloc on the sorted dict keys.
    # to avoid creating a temporary area column (which doesn't work for GeoSeries).
    length_mapper = dict(enumerate(gdf.length.values))
    sorted_lengths = dict(sorted(length_mapper.items(), key=_get_dict_value))
    return gdf.iloc[list(sorted_lengths)]


def sort_small_first(gdf: GeoDataFrame | GeoSeries) -> GeoDataFrame | GeoSeries:
    """Sort GeoDataFrame by area in ascending order.

    Args:
        gdf: A GeoDataFrame or GeoSeries.

    Returns:
        A GeoDataFrame or GeoSeries sorted from small to large in area.

    """
    # using enumerate, then iloc on the sorted dict keys.
    # to avoid creating a temporary area column (which doesn't work for GeoSeries).
    area_mapper = dict(enumerate(gdf.area.values))
    sorted_areas = dict(sorted(area_mapper.items(), key=_get_dict_value))
    return gdf.iloc[list(sorted_areas)]


def _get_dict_value(item: tuple[Hashable, Any]) -> Any:
    return item[1]


def make_lines_between_points(
    *arrs: NDArray[Point] | GeometryArray | GeoSeries,
) -> NDArray[LineString]:
    """Creates an array of linestrings from two or more arrays of points.

    The lines are created rowwise, meaning from arr0[0] to arr1[0], from arr0[1] to arr1[1]...
    If more than two arrays are passed, e.g. three arrays,
    the lines will go from arr0[0] via arr1[0] to arr2[0].

    Args:
        arrs: 1 dimensional arrays of point geometries.
            All arrays must have the same shape.
            Must be at least two arrays.

    Returns:
        A numpy array of linestrings.

    """
    coords = [get_coordinates(arr, return_index=False) for arr in arrs]
    return linestrings(
        np.concatenate([coords_arr[:, None, :] for coords_arr in coords], axis=1)
    )


def random_points(n: int, loc: float | int = 0.5) -> GeoDataFrame:
    """Creates a GeoDataFrame with n random points.

    Args:
        n: Number of points to create.
        loc: Mean ('centre') of the distribution.

    Returns:
        A GeoDataFrame of points with n rows.

    Examples:
    ---------
    >>> import sgis as sg
    >>> points = sg.random_points(10_000)
    >>> points
                         geometry
    0     POINT (0.62044 0.22805)
    1     POINT (0.31885 0.38109)
    2     POINT (0.39632 0.61130)
    3     POINT (0.99401 0.35732)
    4     POINT (0.76403 0.73539)
    ...                       ...
    9995  POINT (0.90433 0.75080)
    9996  POINT (0.10959 0.59785)
    9997  POINT (0.00330 0.79168)
    9998  POINT (0.90926 0.96215)
    9999  POINT (0.01386 0.22935)
    [10000 rows x 1 columns]

    Values with a mean of 100.

    >>> points = sg.random_points(10_000, loc=100)
    >>> points
                         geometry
    0      POINT (50.442 199.729)
    1       POINT (26.450 83.367)
    2     POINT (111.054 147.610)
    3      POINT (93.141 141.456)
    4       POINT (94.101 24.837)
    ...                       ...
    9995   POINT (174.344 91.772)
    9996    POINT (95.375 11.391)
    9997    POINT (45.694 60.843)
    9998   POINT (73.261 101.881)
    9999  POINT (134.503 168.155)
    [10000 rows x 1 columns]
    """
    x = np.random.rand(n) * float(loc) * 2
    y = np.random.rand(n) * float(loc) * 2
    return GeoDataFrame(shapely.points(x, y=y), columns=["geometry"])


def random_points_norway(size: int, *, seed: int | None = None) -> GeoDataFrame:
    """Creates a GeoDataFrame with crs=25833 n random points aprox. within the borders of mainland Norway.

    Args:
        size: Number of points to create.
        seed: Optional random seed.

    Returns:
        A GeoDataFrame of points with n rows.
    """
    return random_points_in_polygons(
        [
            shapely.wkt.loads(x)
            for x in [
                "POLYGON ((546192 7586393, 546191 7586393, 526598 7592425, 526597 7592425, 526596 7592425, 526595 7592426, 526594 7592426, 525831 7593004, 525830 7593005, 525327 7593495, 525326 7593496, 525326 7593497, 525325 7593498, 525325 7593499, 525324 7593500, 525192 7594183, 525192 7594184, 524157 7606517, 524157 7606518, 524157 7606519, 524157 7606520, 524157 7606521, 526235 7613535, 526236 7613536, 559423 7676952, 559424 7676953, 559511 7677088, 579978 7708379, 636963 7792940, 636963 7792941, 636964 7792942, 636965 7792943, 641013 7795664, 823514 7912323, 823515 7912323, 823516 7912323, 882519 7931958, 882520 7931959, 882521 7931959, 953896 7939985, 953897 7939985, 973544 7939988, 973545 7939988, 973546 7939988, 975510 7939467, 1051029 7913762, 1051030 7913762, 1055067 7912225, 1055068 7912224, 1056725 7911491, 1098379 7890321, 1098380 7890320, 1098381 7890320, 1099197 7889670, 1099198 7889669, 1099442 7889429, 1099443 7889429, 1099444 7889428, 1099444 7889427, 1099445 7889426, 1099445 7889425, 1099445 7889424, 1099446 7889423, 1114954 7799458, 1115106 7797736, 1115106 7797735, 1115106 7797734, 1115106 7797733, 1115106 7797732, 1115105 7797731, 1115105 7797730, 1114774 7797199, 1112876 7794451, 1057595 7720320, 1057112 7719702, 1057112 7719701, 1057111 7719701, 1057110 7719700, 1057109 7719699, 902599 7637176, 902598 7637176, 902597 7637175, 902596 7637175, 702394 7590633, 702393 7590633, 702392 7590633, 546193 7586393, 546192 7586393))",
                "POLYGON ((60672 6448410, 60671 6448411, 57185 6448783, 39229 6451077, 39228 6451077, 39227 6451077, 27839 6454916, 27838 6454916, 27808 6454929, 27807 6454929, 8939 6465625, 8938 6465626, 7449 6466699, 7448 6466700, 6876 6467215, 6876 6467216, -31966 6512038, -31968 6512040, -32554 6512779, -32554 6512780, -40259 6524877, -42041 6527698, -42217 6528008, -42546 6528677, -42547 6528678, -77251 6614452, -77252 6614453, -77252 6614454, -77252 6614455, -77252 6614456, -77206 6615751, -77206 6615752, -65669 6811422, -65669 6811423, -65608 6812139, -65608 6812140, -65608 6812141, -50907 6879624, -50907 6879625, -50907 6879626, -50906 6879627, -50889 6879658, -50889 6879659, -16217 6934790, -16217 6934791, -16216 6934792, -2958 6949589, -2957 6949590, 55128 6995098, 144915 7064393, 144915 7064394, 144916 7064395, 144958 7064418, 144959 7064418, 144960 7064418, 144961 7064419, 144962 7064419, 144963 7064419, 150493 7064408, 150494 7064408, 150495 7064408, 150770 7064370, 150771 7064370, 150772 7064370, 188559 7048106, 188560 7048105, 188664 7048054, 188665 7048054, 188666 7048053, 357806 6914084, 357807 6914083, 357808 6914082, 357809 6914081, 357809 6914080, 357810 6914079, 357810 6914078, 359829 6906908, 386160 6804356, 386160 6804355, 386160 6804354, 386160 6804353, 386160 6804352, 386160 6804351, 368140 6699014, 368140 6699013, 363725 6675483, 363725 6675482, 361041 6665071, 361040 6665070, 361040 6665069, 308721 6537573, 308720 6537572, 307187 6534433, 307187 6534432, 307186 6534431, 307185 6534430, 307184 6534429, 307183 6534429, 307182 6534428, 303562 6532881, 300420 6531558, 99437 6459510, 99436 6459510, 67654 6449332, 65417 6448682, 65416 6448682, 65415 6448682, 60673 6448410, 60672 6448410))",
                "POLYGON ((219870 6914350, 219869 6914350, 219868 6914351, 219867 6914351, 194827 6928565, 194826 6928566, 193100 6929790, 193099 6929790, 193098 6929791, 193098 6929792, 193097 6929793, 157353 7006877, 157353 7006878, 154402 7017846, 154402 7017847, 154392 7017923, 154392 7017924, 154392 7017925, 154392 7017926, 166616 7077346, 166617 7077347, 169164 7087256, 169165 7087257, 170277 7089848, 173146 7096147, 173147 7096148, 174684 7098179, 174685 7098180, 314514 7253805, 314515 7253805, 314515 7253806, 314516 7253806, 314517 7253807, 314518 7253807, 314519 7253808, 314520 7253808, 314521 7253808, 314522 7253808, 314523 7253808, 314524 7253808, 332374.8847495829 7250200.016409928, 327615 7280207, 327615 7280208, 327615 7280209, 327615 7280210, 328471 7285637, 364549 7480637, 364549 7480638, 367030 7488919, 367030 7488920, 367045 7488948, 367045 7488949, 367046 7488950, 419493 7560257, 472291 7626092, 506326 7665544, 506327 7665545, 506328 7665546, 541847 7692387, 541848 7692388, 541849 7692388, 541850 7692389, 541851 7692389, 541852 7692389, 545852 7692619, 546265 7692617, 546266 7692617, 546267 7692617, 546268 7692617, 546269 7692616, 546270 7692616, 546270 7692615, 546271 7692615, 546272 7692614, 623027 7613734, 623028 7613733, 623029 7613732, 627609 7605928, 627610 7605928, 627610 7605927, 627610 7605926, 627611 7605925, 627611 7605924, 630573 7568363, 630573 7568362, 630573 7568361, 630573 7568360, 630573 7568359, 628567 7562381, 621356 7542293, 621356 7542292, 468368 7221876.188770507, 468368 7221876, 459071 7119021, 459071 7119020, 459071 7119019, 459070 7119018, 459070 7119017, 454728 7109371, 451784 7102984, 449525 7098307, 357809 6914071, 357808 6914070, 357808 6914069, 357807 6914068, 357806 6914068, 357806 6914067, 357805 6914067, 357804 6914066, 353158 6912240, 353157 6912239, 353156 6912239, 351669 6911974, 351668 6911974, 351667 6911974, 219871 6914350, 219870 6914350))",
            ]
        ],
        size=size,
        crs=25833,
        seed=seed,
    ).sample(size)


def random_points_in_polygons(
    polygons: Geometry | GeoDataFrame | GeoSeries,
    size: int,
    *,
    seed: int | None = None,
    crs: Any = 25833,
) -> GeoDataFrame:
    """Creates a GeoDataFrame with n random points within the geometries of 'gdf'.

    Args:
        polygons: A GeoDataFrame or GeoSeries of polygons. Or a single polygon.
        size: Number of points to create.
        seed: Optional random seed.
        crs: Optional crs of the output GeoDataFrame if input is shapely.Geometry.

    Returns:
        A GeoDataFrame of points with n rows.
    """
    if crs is None:
        try:
            crs = polygons.crs
        except AttributeError:
            pass
    rng = np.random.default_rng(seed)
    polygons = to_gdf(polygons, crs=crs).geometry
    bounds = polygons.bounds
    minx = np.repeat(bounds["minx"].values, size)
    maxx = np.repeat(bounds["maxx"].values, size)
    miny = np.repeat(bounds["miny"].values, size)
    maxy = np.repeat(bounds["maxy"].values, size)
    index = np.repeat(np.arange(len(polygons)), size)
    length = len(index)
    out = []
    while sum(len(df) for df in out) < size * len(polygons):
        xs = rng.uniform(low=minx, high=maxx, size=length)
        ys = rng.uniform(low=miny, high=maxy, size=length)
        out.append(
            GeoDataFrame(
                shapely.points(xs, y=ys), index=index, columns=["geometry"], crs=crs
            ).pipe(sfilter, polygons)
        )
    return pd.concat(out).groupby(level=0).sample(size, replace=True).sort_index()


def polygons_to_lines(
    gdf: GeoDataFrame | GeoSeries, copy: bool = True
) -> GeoDataFrame | GeoSeries:
    if not len(gdf):
        return gdf
    if not (gdf.geom_type == "Polygon").all():
        raise ValueError("geometries must be singlepart polygons")
    if copy:
        gdf = gdf.copy()
    geoms = gdf.geometry.values
    exterior_coords, exterior_indices = shapely.get_coordinates(
        shapely.get_exterior_ring(geoms), return_index=True
    )
    exteriors = shapely.linestrings(exterior_coords, indices=exterior_indices)
    max_rings: int = np.max(shapely.get_num_interior_rings(geoms))

    interiors = [
        [LineString(shapely.get_interior_ring(geom, j)) for j in range(max_rings)]
        for i, geom in enumerate(geoms)
    ]

    lines = shapely.union_all(
        np.array(
            [[ext, *int_] for ext, int_ in zip(exteriors, interiors, strict=True)]
        ),
        axis=1,
    )

    gdf.geometry.loc[:] = lines

    return gdf


def to_lines(
    *gdfs: GeoDataFrame, copy: bool = True, split: bool = True
) -> GeoDataFrame:
    """Makes lines out of one or more GeoDataFrames and splits them at intersections.

    The GeoDataFrames' geometries are converted to LineStrings, then unioned together
    and made to singlepart. The lines are split at the intersections. Mimics
    'feature to line' in ArcGIS.

    Args:
        *gdfs: one or more GeoDataFrames.
        copy: whether to take a copy of the incoming GeoDataFrames. Defaults to True.
        split: If True (default), lines will be split at intersections if more than
            one GeoDataFrame is passed as gdfs. Otherwise, a simple concat.

    Returns:
        A GeoDataFrame with singlepart line geometries and columns of all input
            GeoDataFrames.

    Note:
        The index is preserved if only one GeoDataFrame is given, but otherwise
        ignored. This is because the union overlay used if multiple GeoDataFrames
        always ignores the index.

    Examples:
    ---------
    Convert single polygon to linestring.

    >>> import sgis as sg
    >>> from shapely.geometry import Polygon
    >>> poly1 = sg.to_gdf(Polygon([(0, 0), (0, 1), (1, 1), (1, 0)]))
    >>> poly1["poly1"] = 1
    >>> line = sg.to_lines(poly1)
    >>> line
                                                geometry  poly1
    0  LINESTRING (0.00000 0.00000, 0.00000 1.00000, ...      1

    Convert two overlapping polygons to linestrings.

    >>> poly2 = sg.to_gdf(Polygon([(0.5, 0.5), (0.5, 1.5), (1.5, 1.5), (1.5, 0.5)]))
    >>> poly2["poly2"] = 1
    >>> lines = sg.to_lines(poly1, poly2)
    >>> lines
    poly1  poly2                                           geometry
    0    1.0    NaN  LINESTRING (0.00000 0.00000, 0.00000 1.00000, ...
    1    1.0    NaN  LINESTRING (0.50000 1.00000, 1.00000 1.00000, ...
    2    1.0    NaN  LINESTRING (1.00000 0.50000, 1.00000 0.00000, ...
    3    NaN    1.0      LINESTRING (0.50000 0.50000, 0.50000 1.00000)
    4    NaN    1.0  LINESTRING (0.50000 1.00000, 0.50000 1.50000, ...
    5    NaN    1.0      LINESTRING (1.00000 0.50000, 0.50000 0.50000)

    Plot before and after.

    >>> sg.qtm(poly1, poly2)
    >>> lines["l"] = lines.length
    >>> sg.qtm(lines, "l")
    """
    gdf = (
        pd.concat(df.assign(**{"_df_idx": i}) for i, df in enumerate(gdfs))
        .pipe(make_all_singlepart)
        .pipe(clean_geoms)
        .pipe(make_all_singlepart, ignore_index=True)
    )
    geom_col = gdf.geometry.name

    if not len(gdf):
        return gdf.drop(columns="_df_idx")

    geoms = gdf.geometry

    if (geoms.geom_type == "Polygon").all():
        geoms = polygons_to_lines(geoms, copy=copy)
    elif (geoms.geom_type != "LineString").any():
        raise ValueError(
            f"Point geometries not allowed in 'to_lines'. {geoms.geom_type.value_counts()}"
        )

    gdf.loc[:, gdf.geometry.name] = geoms

    if not split:
        return gdf

    out = []
    for i in gdf["_df_idx"].unique():
        these = gdf[gdf["_df_idx"] == i]
        others = gdf.loc[gdf["_df_idx"] != i, [geom_col]]
        intersection_points = these.overlay(others, keep_geom_type=False).explode(
            ignore_index=True
        )
        points = intersection_points[intersection_points.geom_type == "Point"]
        lines = intersection_points[intersection_points.geom_type == "LineString"]
        splitted = _split_lines_by_points_along_line(these, points, splitted_col=None)
        out.append(splitted)
        out.append(lines)

    return (
        pd.concat(out, ignore_index=True)
        .pipe(make_all_singlepart, ignore_index=True)
        .drop(columns="_df_idx")
    )


def _split_lines_by_points_along_line(lines, points, splitted_col: str | None = None):
    precision = 1e-5
    # find the lines that were snapped to (or are very close because of float rounding)
    points_buff = points.buffer(precision, resolution=16).to_frame("geometry")
    relevant_lines, the_other_lines = sfilter_split(lines, points_buff)

    if not len(relevant_lines):
        if splitted_col:
            return lines.assign(**{splitted_col: 0})
        return lines

    # need consistent coordinate dimensions later
    # (doing it down here to not overwrite the original data)
    relevant_lines.geometry = shapely.force_2d(relevant_lines.geometry)
    points.geometry = shapely.force_2d(points.geometry)

    # split the lines with buffer + difference, since shaply.split usually doesn't work
    # relevant_lines["_idx"] = range(len(relevant_lines))
    splitted = relevant_lines.overlay(points_buff, how="difference").explode(
        ignore_index=True
    )

    # linearrings (maybe coded as linestrings) that were not split,
    # do not have edges and must be added in the end
    boundaries = splitted.geometry.boundary
    circles = splitted[boundaries.is_empty]
    splitted = splitted[~boundaries.is_empty]

    if not len(splitted):
        return pd.concat([the_other_lines, circles], ignore_index=True)

    # the endpoints of the new lines are now sligtly off. Using get_k_nearest_neighbors
    # to get the exact snapped point coordinates, . This will map the sligtly
    # wrong line endpoints with the point the line was split by.

    points["point_coords"] = [(geom.x, geom.y) for geom in points.geometry]

    # get line endpoints as columns (source_coords and target_coords)
    splitted = make_edge_coords_cols(splitted)

    splitted_source = to_gdf(splitted["source_coords"], crs=lines.crs)
    splitted_target = to_gdf(splitted["target_coords"], crs=lines.crs)

    def get_nearest(splitted: GeoDataFrame, points: GeoDataFrame) -> pd.DataFrame:
        """Find the nearest snapped point for each source and target of the lines."""
        return get_k_nearest_neighbors(splitted, points, k=1).loc[
            lambda x: x["distance"] <= precision * 2
        ]

    # points = points.set_index("point_coords")
    points.index = points.geometry
    dists_source = get_nearest(splitted_source, points)
    dists_target = get_nearest(splitted_target, points)

    # neighbor_index: point coordinates as tuple
    pointmapper_source: pd.Series = dists_source["neighbor_index"]
    pointmapper_target: pd.Series = dists_target["neighbor_index"]

    # now, we can replace the source/target coordinate with the coordinates of
    # the snapped points.

    splitted = _change_line_endpoint(
        splitted,
        indices=dists_source.index,
        pointmapper=pointmapper_source,
        change_what="first",
    )

    # same for the lines where the target was split, but change the last coordinate
    splitted = _change_line_endpoint(
        splitted,
        indices=dists_target.index,
        pointmapper=pointmapper_target,
        change_what="last",
    )

    if splitted_col:
        splitted[splitted_col] = 1

    return pd.concat([the_other_lines, splitted, circles], ignore_index=True).drop(
        ["source_coords", "target_coords"], axis=1
    )


def _change_line_endpoint(
    gdf: GeoDataFrame,
    indices: pd.Index,
    pointmapper: pd.Series,
    change_what: str | int,
) -> GeoDataFrame:
    """Modify the endpoints of selected lines in a GeoDataFrame based on an index mapping.

    This function updates the geometry of specified line features within a GeoDataFrame,
    changing either the first or last point of each line to new coordinates provided by a mapping.
    It is typically used in scenarios where line endpoints need to be adjusted to new locations,
    such as in network adjustments or data corrections.

    Args:
        gdf: A GeoDataFrame containing line geometries.
        indices: An Index object identifying the rows in the GeoDataFrame whose endpoints will be changed.
        pointmapper: A Series mapping from the index of lines to new point geometries.
        change_what: Specifies which endpoint of the line to change. Accepts 'first' or 0 for the
            starting point, and 'last' or -1 for the ending point.

    Returns:
        A GeoDataFrame with the specified line endpoints updated according to the pointmapper.

    Raises:
        ValueError: If `change_what` is not one of the accepted values ('first', 'last', 0, -1).
    """
    assert gdf.index.is_unique

    if change_what == "first" or change_what == 0:
        keep = "first"
    elif change_what == "last" or change_what == -1:
        keep = "last"
    else:
        raise ValueError(
            f"change_what should be 'first' or 'last' or 0 or -1. Got {change_what}"
        )

    is_relevant = gdf.index.isin(indices)
    relevant_lines = gdf.loc[is_relevant]

    relevant_lines.geometry = extract_unique_points(relevant_lines.geometry)
    relevant_lines = relevant_lines.explode(index_parts=False)

    relevant_lines.loc[lambda x: ~x.index.duplicated(keep=keep), "geometry"] = (
        relevant_lines.loc[lambda x: ~x.index.duplicated(keep=keep)]
        .index.map(pointmapper)
        .values
    )

    is_line = relevant_lines.groupby(level=0).size() > 1
    relevant_lines_mapped = (
        relevant_lines.loc[is_line].groupby(level=0)["geometry"].agg(LineString)
    )

    gdf.loc[relevant_lines_mapped.index, "geometry"] = relevant_lines_mapped

    return gdf


def make_edge_coords_cols(gdf: GeoDataFrame) -> GeoDataFrame:
    """Get the wkt of the first and last points of lines as columns.

    It takes a GeoDataFrame of LineStrings and returns a GeoDataFrame with two new
    columns, source_coords and target_coords, which are the x and y coordinates of the
    first and last points of the LineStrings in a tuple. The lines all have to be

    Args:
        gdf (GeoDataFrame): the GeoDataFrame with the lines

    Returns:
        A GeoDataFrame with new columns 'source_coords' and 'target_coords'
    """
    try:
        gdf, endpoints = _prepare_make_edge_cols_simple(gdf)
    except ValueError:
        gdf, endpoints = _prepare_make_edge_cols(gdf)

    coords = [(geom.x, geom.y) for geom in endpoints.geometry]
    gdf["source_coords"], gdf["target_coords"] = (
        coords[0::2],
        coords[1::2],
    )

    return gdf


def make_edge_wkt_cols(gdf: GeoDataFrame) -> GeoDataFrame:
    """Get coordinate tuples of the first and last points of lines as columns.

    It takes a GeoDataFrame of LineStrings and returns a GeoDataFrame with two new
    columns, source_wkt and target_wkt, which are the WKT representations of the first
    and last points of the LineStrings

    Args:
        gdf (GeoDataFrame): the GeoDataFrame with the lines

    Returns:
        A GeoDataFrame with new columns 'source_wkt' and 'target_wkt'
    """
    try:
        gdf, endpoints = _prepare_make_edge_cols_simple(gdf)
    except ValueError:
        gdf, endpoints = _prepare_make_edge_cols(gdf)

    wkt_geom = [
        f"POINT ({x} {y})" for x, y in zip(endpoints.x, endpoints.y, strict=True)
    ]
    gdf["source_wkt"], gdf["target_wkt"] = (
        wkt_geom[0::2],
        wkt_geom[1::2],
    )

    return gdf


def _prepare_make_edge_cols(lines: GeoDataFrame) -> tuple[GeoDataFrame, GeoDataFrame]:

    lines = lines.loc[lines.geom_type != "LinearRing"]

    if not (lines.geom_type == "LineString").all():
        multilinestring_error_message = (
            "MultiLineStrings have more than two endpoints. "
            "Try shapely.line_merge and/or explode() to get LineStrings. "
            "Or use the Network class methods, where the lines are prepared correctly."
        )
        if (lines.geom_type == "MultiLinestring").any():
            raise ValueError(multilinestring_error_message)
        else:
            raise ValueError(
                "You have mixed geometries. Only lines are accepted. "
                "Try using: to_single_geom_type(gdf, 'lines')."
            )

    geom_col = lines._geometry_column_name

    # some LineStrings are in fact rings and must be removed manually
    lines, _ = split_out_circles(lines)

    endpoints = lines[geom_col].boundary.explode(ignore_index=True)

    if len(lines) and len(endpoints) / len(lines) != 2:
        raise ValueError(
            "The lines should have only two endpoints each. "
            "Try splitting multilinestrings with explode.",
            lines[geom_col],
        )

    return lines, endpoints


def _prepare_make_edge_cols_simple(
    lines: GeoDataFrame,
) -> tuple[GeoDataFrame, GeoDataFrame]:
    """Faster version of _prepare_make_edge_cols."""
    endpoints = lines[lines._geometry_column_name].boundary.explode(ignore_index=True)

    if len(lines) and len(endpoints) / len(lines) != 2:
        raise ValueError(
            "The lines should have only two endpoints each. "
            "Try splitting multilinestrings with explode."
        )

    return lines, endpoints


def clean_clip(
    gdf: GeoDataFrame | GeoSeries,
    mask: GeoDataFrame | GeoSeries | Geometry,
    keep_geom_type: bool | None = None,
    geom_type: str | None = None,
    **kwargs,
) -> GeoDataFrame | GeoSeries:
    """Clips and clean geometries.

    Geopandas.clip does a "fast and dirty clipping, with no guarantee for valid
    outputs". Here, the clipped geometries are made valid, and empty and NaN
    geometries are removed.

    Args:
        gdf: GeoDataFrame or GeoSeries to be clipped
        mask: the geometry to clip gdf
        geom_type: Optionally specify what geometry type to keep.,
            if there are mixed geometry types. Must be either "polygon",
            "line" or "point".
        keep_geom_type: Defaults to None, meaning True if 'geom_type' is given
            and True if the geometries are single-typed and False if the geometries
            are mixed.
        **kwargs: Keyword arguments passed to geopandas.GeoDataFrame.clip

    Returns:
        The cleanly clipped GeoDataFrame.

    Raises:
        TypeError: If gdf is not of type GeoDataFrame or GeoSeries.
    """
    if not isinstance(gdf, (GeoDataFrame, GeoSeries)):
        raise TypeError(f"'gdf' should be GeoDataFrame or GeoSeries, got {type(gdf)}")

    gdf, geom_type, keep_geom_type = _determine_geom_type_args(
        gdf, geom_type, keep_geom_type
    )

    try:
        gdf = gdf.clip(mask, **kwargs).pipe(clean_geoms)
    except Exception:
        gdf = clean_geoms(gdf)
        try:
            mask = clean_geoms(mask)
        except TypeError:
            mask = make_valid(mask)

        return gdf.clip(mask, **kwargs).pipe(clean_geoms)

    if keep_geom_type:
        gdf = to_single_geom_type(gdf, geom_type)

    return gdf


def split_out_circles(
    lines: GeoDataFrame | GeoSeries,
) -> tuple[GeoDataFrame | GeoSeries, GeoDataFrame | GeoSeries]:
    boundaries = lines.geometry.boundary
    is_circle = (~boundaries.is_empty).values
    return lines.iloc[is_circle], lines.iloc[~is_circle]


def extend_lines(arr1, arr2, distance) -> NDArray[LineString]:
    if len(arr1) != len(arr2):
        raise ValueError
    if not len(arr1):
        return arr1

    arr1, arr2 = arr2, arr1  # TODO fix

    coords1 = coordinate_array(arr1)
    coords2 = coordinate_array(arr2)

    dx = coords2[:, 0] - coords1[:, 0]
    dy = coords2[:, 1] - coords1[:, 1]
    len_xy = np.sqrt((dx**2.0) + (dy**2.0))
    x = coords1[:, 0] + (coords1[:, 0] - coords2[:, 0]) / len_xy * distance
    y = coords1[:, 1] + (coords1[:, 1] - coords2[:, 1]) / len_xy * distance

    new_points = np.array([None for _ in range(len(arr1))])
    new_points[~np.isnan(x)] = shapely.points(x[~np.isnan(x)], y[~np.isnan(x)])

    new_points[~np.isnan(x)] = make_lines_between_points(
        arr2[~np.isnan(x)], new_points[~np.isnan(x)]
    )
    return new_points


def multipoints_to_line_segments_numpy(
    points: GeoSeries | NDArray[MultiPoint] | MultiPoint,
    cycle: bool = False,
) -> list[LineString]:
    try:
        arr = get_parts(points.geometry.values)
    except AttributeError:
        arr = get_parts(points)

    line_between_last_and_first = [LineString([arr[-1], arr[0]])] if cycle else []
    return [
        LineString([p0, p1]) for p0, p1 in itertools.pairwise(arr)
    ] + line_between_last_and_first


def multipoints_to_line_segments(
    multipoints: GeoSeries | GeoDataFrame, cycle: bool = True  # to_next: bool = True,
) -> GeoSeries | GeoDataFrame:

    if not len(multipoints):
        return multipoints

    if isinstance(multipoints, GeoDataFrame):
        df = multipoints.drop(columns=multipoints.geometry.name)
        multipoints = multipoints.geometry
        was_gdf = True
    else:
        multipoints = to_geoseries(multipoints)
        was_gdf = False

    multipoints = to_geoseries(multipoints)

    segs = pd.Series(
        [
            multipoints_to_line_segments_numpy(geoms, cycle=cycle)
            for geoms in multipoints
        ],
        index=multipoints.index,
    ).explode()

    segs = GeoSeries(segs, crs=multipoints.crs, name=multipoints.name)

    if was_gdf:
        return GeoDataFrame(df.join(segs), geometry=segs.name, crs=segs.crs)
    else:
        return segs


def get_line_segments(
    lines: GeoDataFrame | GeoSeries, extract_unique: bool = False, cycle=False
) -> GeoDataFrame:
    try:
        assert lines.index.is_unique
    except AttributeError:
        pass

    if isinstance(lines, GeoDataFrame):
        df = lines.drop(columns=lines.geometry.name)
        lines = lines.geometry
        was_gdf = True
    else:
        lines = to_geoseries(lines)
        was_gdf = False

    partial_segs_func = functools.partial(
        multipoints_to_line_segments_numpy, cycle=cycle
    )
    if extract_unique:
        points = extract_unique_points(lines.geometry.values)
        segs = pd.Series(
            [partial_segs_func(geoms) for geoms in points],
            index=lines.index,
        ).explode()
    else:
        coords, indices = shapely.get_coordinates(lines, return_index=True)
        points = GeoSeries(shapely.points(coords), index=indices)
        index_mapper = {
            i: idx
            for i, idx in zip(
                np.unique(indices), lines.index.drop_duplicates(), strict=True
            )
        }
        points.index = points.index.map(index_mapper)

        segs = points.groupby(level=0).agg(partial_segs_func).explode()
    segs = GeoSeries(segs, crs=lines.crs, name=lines.name)

    if was_gdf:
        return GeoDataFrame(df.join(segs), geometry=segs.name, crs=lines.crs)
    else:
        return segs


def get_index_right_columns(gdf: pd.DataFrame | pd.Series) -> list[str]:
    """Get a list of what will be the resulting columns in an sjoin."""
    if gdf.index.name is None and all(name is None for name in gdf.index.names):
        if gdf.index.nlevels == 1:
            return ["index_right"]
        else:
            return [f"index_right{i}" for i in range(gdf.index.nlevels)]
    else:
        return gdf.index.names


def points_in_bounds(
    gdf: GeoDataFrame | GeoSeries, gridsize: int | float
) -> GeoDataFrame:
    """Get a GeoDataFrame of points within the bounds of the GeoDataFrame."""
    minx, miny, maxx, maxy = to_bbox(gdf)
    try:
        crs = gdf.crs
    except AttributeError:
        crs = None

    xs = np.linspace(minx, maxx, num=int((maxx - minx) / gridsize))
    ys = np.linspace(miny, maxy, num=int((maxy - miny) / gridsize))
    x_coords, y_coords = np.meshgrid(xs, ys, indexing="ij")
    coords = np.concatenate((x_coords.reshape(-1, 1), y_coords.reshape(-1, 1)), axis=1)
    return to_gdf(coords, crs=crs)


def points_in_polygons(
    gdf: GeoDataFrame | GeoSeries, gridsize: int | float
) -> GeoDataFrame:
    index_right_col = get_index_right_columns(gdf)
    out = points_in_bounds(gdf, gridsize).sjoin(gdf).set_index(index_right_col)
    out.index.name = gdf.index.name
    return out.sort_index()


def _determine_geom_type_args(
    gdf: GeoDataFrame, geom_type: str | None, keep_geom_type: bool | None
) -> tuple[GeoDataFrame, str, bool]:
    if geom_type:
        gdf = to_single_geom_type(gdf, geom_type)
        keep_geom_type = True
    elif keep_geom_type is None:
        geom_type = get_geom_type(gdf)
        if geom_type == "mixed":
            keep_geom_type = False
        else:
            keep_geom_type = True
    elif keep_geom_type:
        geom_type = get_geom_type(gdf)
        if geom_type == "mixed":
            raise ValueError("Cannot set keep_geom_type=True with mixed geometries")
    return gdf, geom_type, keep_geom_type



"""Functions that buffer, dissolve and/or explodes geometries while fixing geometries.

The functions do the same as the geopandas buffer, dissolve and explode methods, except
for the following:

- Geometries are made valid after buffer and dissolve.

- The buffer resolution defaults to 30 (geopandas' default is 16).

- If 'by' is not specified, the index will be labeled 0, 1, …, n - 1 after exploded, instead of 0, 0, …, 0 as it will with the geopandas defaults.

- index_parts is set to False, which will be the default in a future version of geopandas.

- The buff function returns a GeoDataFrame, the geopandas method returns a GeoSeries.
"""

from collections.abc import Callable
from collections.abc import Sequence

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from shapely import get_num_geometries

from ..conf import _get_instance
from ..conf import config
from ..parallel.parallel import Parallel
from .geometry_types import make_all_singlepart
from .polygon_operations import get_cluster_mapper
from .polygon_operations import get_grouped_centroids
from .runners import UnionRunner
from .utils import _unary_union_for_notna


def _decide_ignore_index(kwargs: dict) -> tuple[dict, bool]:
    if "ignore_index" in kwargs:
        ignore_index = kwargs.pop("ignore_index")
        return kwargs, ignore_index

    if kwargs.get("by", None) is None:
        return kwargs, True

    if kwargs.get("as_index", True):
        return kwargs, False

    return kwargs, True


def buffdissexp(
    gdf: GeoDataFrame,
    distance: int | float,
    *,
    resolution: int = 30,
    index_parts: bool = False,
    copy: bool = True,
    grid_size: float | int | None = None,
    join_style: int | str = "round",
    n_jobs: int = 1,
    **dissolve_kwargs,
) -> GeoDataFrame:
    """Buffers and dissolves overlapping geometries.

    It takes a GeoDataFrame and buffer, fixes, dissolves, fixes and explodes geometries.
    If the 'by' parameter is not specified, the index will labeled 0, 1, …, n - 1,
    instead of 0, 0, …, 0. If 'by' is speficied, this will be the index.

    Args:
        gdf: the GeoDataFrame that will be buffered, dissolved and exploded.
        distance: the distance (meters, degrees, depending on the crs) to buffer
            the geometry by
        resolution: The number of segments used to approximate a quarter circle.
            Here defaults to 30, as opposed to the default 16 in geopandas.
        index_parts: If False (default), the index after dissolve is respected. If
            True, an integer index level is added during explode.
        copy: Whether to copy the GeoDataFrame before buffering. Defaults to True.
        grid_size: Rounding of the coordinates. Defaults to None.
        n_jobs: Number of threads to use. Defaults to 1.
        join_style: Buffer join style.
        **dissolve_kwargs: additional keyword arguments passed to geopandas' dissolve.

    Returns:
        A buffered GeoDataFrame where overlapping geometries are dissolved.
    """
    dissolve_kwargs, ignore_index = _decide_ignore_index(dissolve_kwargs)

    dissolved = buffdiss(
        gdf,
        distance,
        resolution=resolution,
        copy=copy,
        grid_size=grid_size,
        n_jobs=n_jobs,
        join_style=join_style,
        **dissolve_kwargs,
    )

    return make_all_singlepart(
        dissolved, ignore_index=ignore_index, index_parts=index_parts
    )


def buffdiss(
    gdf: GeoDataFrame,
    distance: int | float,
    resolution: int = 30,
    copy: bool = True,
    n_jobs: int = 1,
    join_style: int | str = "round",
    **dissolve_kwargs,
) -> GeoDataFrame:
    """Buffers and dissolves geometries.

    It takes a GeoDataFrame and buffer, fixes, dissolves and fixes geometries.
    If the 'by' parameter is not specified, the index will labeled 0, 1, …, n - 1,
    instead of 0, 0, …, 0. If 'by' is speficied, this will be the index.

    Args:
        gdf: the GeoDataFrame that will be
            buffered and dissolved.
        distance: the distance (meters, degrees, depending on the crs) to buffer
            the geometry by
        resolution: The number of segments used to approximate a quarter circle.
            Here defaults to 30, as opposed to the default 16 in geopandas.
        join_style: Buffer join style.
        copy: Whether to copy the GeoDataFrame before buffering. Defaults to True.
        n_jobs: Number of threads to use. Defaults to 1.
        **dissolve_kwargs: additional keyword arguments passed to geopandas' dissolve.

    Returns:
        A buffered GeoDataFrame where geometries are dissolved.

    Examples:
    ---------
    Create some random points.

    >>> import sgis as sg
    >>> import numpy as np
    >>> points = sg.read_parquet_url(
    ...     "https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/points_oslo.parquet"
    ... )[["geometry"]]
    >>> points["group"] = np.random.choice([*"abd"], len(points))
    >>> points["number"] = np.random.random(size=len(points))
    >>> points
                               geometry group    number
    0    POINT (263122.700 6651184.900)     a  0.878158
    1    POINT (272456.100 6653369.500)     a  0.693311
    2    POINT (270082.300 6653032.700)     b  0.323960
    3    POINT (259804.800 6650339.700)     a  0.606745
    4    POINT (272876.200 6652889.100)     a  0.194360
    ..                              ...   ...       ...
    995  POINT (266801.700 6647844.500)     a  0.814424
    996  POINT (261274.000 6653593.400)     b  0.769479
    997  POINT (263542.900 6645427.000)     a  0.925991
    998  POINT (269226.700 6650628.000)     b  0.431972
    999  POINT (264570.300 6644239.500)     d  0.555239

    Buffer by 100 meters and dissolve.

    >>> sg.buffdiss(points, 100)
                                                geometry group    number
    0  MULTIPOLYGON (((256421.833 6649878.117, 256420...     d  0.580157

    Dissolve by 'group' and get sum of columns.

    >>> sg.buffdiss(points, 100, by="group", aggfunc="sum")
                                                    geometry      number
    group
    a      MULTIPOLYGON (((258866.258 6648220.031, 258865...  167.265619
    b      MULTIPOLYGON (((258404.858 6647830.931, 258404...  171.939169
    d      MULTIPOLYGON (((258180.258 6647935.731, 258179...  156.964300

    To get the 'by' columns as columns, not index.

    >>> sg.buffdiss(points, 100, by="group", as_index=False)
      group                                           geometry    number
    0     a  MULTIPOLYGON (((258866.258 6648220.031, 258865...  0.323948
    1     b  MULTIPOLYGON (((258404.858 6647830.931, 258404...  0.687635
    2     d  MULTIPOLYGON (((258180.258 6647935.731, 258179...  0.580157
    """
    buffered = buff(
        gdf, distance, resolution=resolution, copy=copy, join_style=join_style
    )

    return _dissolve(buffered, n_jobs=n_jobs, **dissolve_kwargs)


def _dissolve(
    gdf: GeoDataFrame,
    aggfunc: str = "first",
    grid_size: None | float = None,
    as_index: bool = True,
    n_jobs: int = 1,
    union_runner: UnionRunner | None = None,
    **dissolve_kwargs,
) -> GeoDataFrame:
    if not len(gdf):
        return gdf

    if union_runner is None:
        union_runner = _get_instance(config, "union_runner", n_jobs=n_jobs)

    geom_col = gdf.geometry.name
    by = dissolve_kwargs.pop("by", None)
    by_was_none = not bool(by)

    # make sure geometries are dissolved rowwise to make dissolving simpler later
    more_than_one = get_num_geometries(gdf.geometry.values) > 1
    gdf.loc[more_than_one, geom_col] = gdf.loc[more_than_one, geom_col].apply(
        _unary_union_for_notna
    )

    if by is None and dissolve_kwargs.get("level") is None:
        by = np.zeros(len(gdf), dtype="int64")
        other_cols = list(gdf.columns.difference({geom_col}))
    else:
        if isinstance(by, str):
            by = [by]
        other_cols = list(gdf.columns.difference({geom_col} | set(by or {})))
    try:
        is_one_hit = (
            gdf.groupby(by, as_index=True, **dissolve_kwargs).transform("size") == 1
        )
    except IndexError:
        # if no rows after dropping na if dropna=True
        original_by = [x for x in by]
        query = gdf[by.pop(0)].notna()
        for col in gdf[by]:
            query &= gdf[col].notna()
        gdf = gdf.loc[query]
        assert not len(gdf), gdf
        if as_index and not by_was_none:
            gdf = gdf.set_index(original_by)
        return gdf

    if as_index and not by_was_none:
        one_hit = gdf[is_one_hit].set_index(by)
    else:
        one_hit = gdf[is_one_hit]
    many_hits = gdf[~is_one_hit]

    if not len(many_hits):
        return GeoDataFrame(one_hit, geometry=geom_col, crs=gdf.crs)

    dissolved = many_hits.groupby(by, as_index=True, **dissolve_kwargs)[other_cols].agg(
        aggfunc
    )
    dissolved[geom_col] = union_runner.run(
        many_hits,
        by=by,
        grid_size=grid_size,
        as_index=True,
        **dissolve_kwargs,
    )
    if not as_index:
        dissolved = dissolved.reset_index()
    try:
        return GeoDataFrame(
            pd.concat([dissolved, one_hit]).sort_index(), geometry=geom_col, crs=gdf.crs
        )
    except TypeError as e:
        raise e.__class__(f"{e}. {dissolved.index}. {one_hit.index}") from e


def diss(
    gdf: GeoDataFrame,
    by: str | Sequence[str] | None = None,
    aggfunc: str | Callable | dict[str, str | Callable] = "first",
    as_index: bool = True,
    grid_size: float | int | None = None,
    n_jobs: int = 1,
    **dissolve_kwargs,
) -> GeoDataFrame:
    """Dissolves geometries.

    It takes a GeoDataFrame and dissolves and fixes geometries.

    Args:
        gdf: the GeoDataFrame that will be dissolved and exploded.
        by: Columns to dissolve by.
        aggfunc: How to aggregate the non-geometry colums not in "by".
        as_index: Whether the 'by' columns should be returned as index. Defaults to
            True to be consistent with geopandas.
        grid_size: Rounding of the coordinates. Defaults to None.
        n_jobs: Number of threads to use. Defaults to 1.
        **dissolve_kwargs: additional keyword arguments passed to geopandas' dissolve.

    Returns:
        A GeoDataFrame with dissolved geometries.
    """
    if not len(gdf):
        if as_index:
            try:
                return gdf.set_index(by)
            except Exception:
                return gdf
        else:
            return gdf

    return _dissolve(
        gdf,
        by=by,
        aggfunc=aggfunc,
        grid_size=grid_size,
        n_jobs=n_jobs,
        as_index=as_index,
        **dissolve_kwargs,
    )


def dissexp(
    gdf: GeoDataFrame,
    by: str | Sequence[str] | None = None,
    aggfunc: str | Callable | dict[str, str | Callable] = "first",
    as_index: bool = True,
    index_parts: bool = False,
    grid_size: float | int | None = None,
    n_jobs: int = 1,
    **dissolve_kwargs,
) -> GeoDataFrame:
    """Dissolves overlapping geometries.

    It takes a GeoDataFrame and dissolves, fixes and explodes geometries.

    Args:
        gdf: the GeoDataFrame that will be dissolved and exploded.
        by: Columns to dissolve by.
        aggfunc: How to aggregate the non-geometry colums not in "by".
        as_index: Whether the 'by' columns should be returned as index. Defaults to
            True to be consistent with geopandas.
        index_parts: If False (default), the index after dissolve is respected. If
            True, an integer index level is added during explode.
        grid_size: Rounding of the coordinates. Defaults to None.
        n_jobs: Number of threads to use. Defaults to 1.
        **dissolve_kwargs: additional keyword arguments passed to geopandas' dissolve.

    Returns:
        A GeoDataFrame where overlapping geometries are dissolved.
    """
    dissolve_kwargs = dissolve_kwargs | {
        "by": by,
        "as_index": as_index,
    }

    dissolve_kwargs, ignore_index = _decide_ignore_index(dissolve_kwargs)

    dissolved = diss(
        gdf, aggfunc=aggfunc, grid_size=grid_size, n_jobs=n_jobs, **dissolve_kwargs
    )

    return make_all_singlepart(
        dissolved, ignore_index=ignore_index, index_parts=index_parts
    )


def dissexp_by_cluster(
    gdf: GeoDataFrame,
    predicate: str | None = "intersects",
    n_jobs: int = 1,
    **dissolve_kwargs,
) -> GeoDataFrame:
    """Dissolves overlapping geometries through clustering with sjoin and networkx.

    Works exactly like dissexp, but, before dissolving, the geometries are divided
    into clusters based on overlap (uses the function sgis.get_polygon_clusters).
    The geometries are then dissolved based on this column (and optionally other
    columns).

    This might be many times faster than a regular dissexp, if there are many
    non-overlapping geometries.

    Args:
        gdf: the GeoDataFrame that will be dissolved and exploded.
        predicate: Spatial predicate to use.
        n_jobs: Number of threads to use. Defaults to 1.
        **dissolve_kwargs: Keyword arguments passed to geopandas' dissolve.

    Returns:
        A GeoDataFrame where overlapping geometries are dissolved.
    """
    return _run_func_by_cluster(
        dissexp, gdf, predicate=predicate, n_jobs=n_jobs, **dissolve_kwargs
    )


def diss_by_cluster(
    gdf: GeoDataFrame, predicate=None, n_jobs: int = 1, **dissolve_kwargs
) -> GeoDataFrame:
    """Dissolves overlapping geometries through clustering with sjoin and networkx.

    Works exactly like dissexp, but, before dissolving, the geometries are divided
    into clusters based on overlap (uses the function sgis.get_polygon_clusters).
    The geometries are then dissolved based on this column (and optionally other
    columns).

    This might be many times faster than a regular dissexp, if there are many
    non-overlapping geometries.

    Args:
        gdf: the GeoDataFrame that will be dissolved and exploded.
        predicate: Spatial predicate to use.
        n_jobs: Number of threads to use. Defaults to 1.
        **dissolve_kwargs: Keyword arguments passed to geopandas' dissolve.

    Returns:
        A GeoDataFrame where overlapping geometries are dissolved.
    """
    return _run_func_by_cluster(
        diss, gdf, predicate=predicate, n_jobs=n_jobs, **dissolve_kwargs
    )


def _run_func_by_cluster(
    func: Callable,
    gdf: GeoDataFrame,
    predicate: str | None = "intersects",
    n_jobs: int = 1,
    **dissolve_kwargs,
) -> GeoDataFrame:
    is_geoseries = isinstance(gdf, GeoSeries)

    processes = dissolve_kwargs.pop("processes", 1)

    by = dissolve_kwargs.pop("by", [])
    if isinstance(by, str):
        by = [by]
    elif by:
        by = list(by)

    if not len(gdf):
        return func(gdf, by=by, **dissolve_kwargs)

    def get_group_clusters(group: GeoDataFrame):
        """Adds cluster column. Applied to each group because much faster."""
        group = group[~group.is_empty].reset_index(drop=True)
        group["_cluster"] = get_cluster_mapper(group, predicate=predicate)
        group["_cluster"] = get_grouped_centroids(group, groupby="_cluster")
        return group

    gdf = make_all_singlepart(gdf)

    if by:
        if processes == 1:
            gdf = gdf.groupby(by, group_keys=False, dropna=False, as_index=False).apply(
                get_group_clusters
            )
        else:
            gdf = pd.concat(
                Parallel(processes, backend="loky").map(
                    get_group_clusters,
                    [
                        gdf[lambda x: x[by].values == values]
                        for values in np.unique(gdf[by].values)
                    ],
                ),
            )
        _by = ["_cluster"] + by
    else:
        gdf = get_group_clusters(gdf)
        _by = ["_cluster"]

    if processes == 1:
        dissolved = func(gdf, by=_by, n_jobs=n_jobs, **dissolve_kwargs)
    else:
        dissolved = pd.concat(
            Parallel(processes, backend="loky").map(
                func,
                [
                    gdf[gdf["_cluster"] == cluster]
                    for cluster in gdf["_cluster"].unique()
                ],
                kwargs=dissolve_kwargs | {"n_jobs": n_jobs, "by": _by},
            ),
        )

    if not by:
        dissolved = dissolved.reset_index(drop=True)

    elif dissolve_kwargs.get("as_index", True):
        dissolved.index = dissolved.index.droplevel(0)

    if is_geoseries:
        return dissolved.geometry

    return dissolved.drop("_cluster", axis=1, errors="ignore")


def buffdissexp_by_cluster(
    gdf: GeoDataFrame,
    distance: int | float,
    *,
    resolution: int = 30,
    copy: bool = True,
    n_jobs: int = 1,
    join_style: int | str = "round",
    **dissolve_kwargs,
) -> GeoDataFrame:
    """Buffers and dissolves overlapping geometries.

    Works exactly like buffdissexp, but, before dissolving, the geometries are divided
    into clusters based on overlap (uses the function sgis.get_polygon_clusters).
    The geometries are then dissolved based on this column (and optionally other
    columns).

    This might be many times faster than a regular buffdissexp, if there are many
    non-overlapping geometries.

    Args:
        gdf: the GeoDataFrame that will be buffered, dissolved and exploded.
        distance: the distance (meters, degrees, depending on the crs) to buffer
            the geometry by
        resolution: The number of segments used to approximate a quarter circle.
            Here defaults to 30, as opposed to the default 16 in geopandas.
        join_style: Buffer join style.
        copy: Whether to copy the GeoDataFrame before buffering. Defaults to True.
        n_jobs: int = 1,
        **dissolve_kwargs: additional keyword arguments passed to geopandas' dissolve.

    Returns:
        A buffered GeoDataFrame where overlapping geometries are dissolved.
    """
    buffered = buff(
        gdf,
        distance,
        resolution=resolution,
        copy=copy,
        join_style=join_style,
    )
    return dissexp_by_cluster(buffered, n_jobs=n_jobs, **dissolve_kwargs)


def buff(
    gdf: GeoDataFrame | GeoSeries,
    distance: int | float,
    resolution: int = 30,
    copy: bool = True,
    join_style: int | str = "round",
    **buffer_kwargs,
) -> GeoDataFrame:
    """Buffers a GeoDataFrame with high resolution and returns a new GeoDataFrame.

    Args:
        gdf: the GeoDataFrame that will be buffered, dissolved and exploded.
        distance: the distance (meters, degrees, depending on the crs) to buffer
            the geometry by
        resolution: The number of segments used to approximate a quarter circle.
            Here defaults to 30, as opposed to the default 16 in geopandas.
        join_style: Buffer join style.
        copy: Whether to copy the GeoDataFrame before buffering. Defaults to True.
        **buffer_kwargs: additional keyword arguments passed to geopandas' buffer.

    Returns:
        A buffered GeoDataFrame.
    """
    if isinstance(gdf, GeoSeries):
        return gdf.buffer(
            distance, resolution=resolution, join_style=join_style, **buffer_kwargs
        ).make_valid()

    if copy:
        gdf = gdf.copy()

    gdf[gdf.geometry.name] = gdf.buffer(
        distance, resolution=resolution, join_style=join_style, **buffer_kwargs
    ).make_valid()

    return gdf



from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from typing import Any

try:
    from dapla import FileClient
    from dapla.gcs import GCSFileSystem
except ImportError:

    class GCSFileSystem:  # type: ignore[no-redef]
        """Placeholder."""


from ._is_dapla import is_dapla


@contextmanager
def opener(
    path, mode: str = "rb", file_system: GCSFileSystem | None = None
) -> Generator[str | Any, None, None]:
    """Yields a gcs buffer if in Dapla, otherwise yields the path.

    Example:
    -------
    >>> with opener(path) as file:
    >>>     with rasterio.open(file) as src:
    >>>         array = src.read()
    """
    if is_dapla():
        if file_system is None:
            file_system = FileClient.get_gcs_file_system()
        yield file_system.open(str(path), mode=mode)
    else:
        yield str(path)



from xyzservices import Bunch
from xyzservices import TileProvider
from xyzservices import providers

kartverket = Bunch(
    topo=TileProvider(
        name="Topografisk norgeskart",
        url="https://cache.kartverket.no/v1/wmts/1.0.0/topo/default/webmercator/{z}/{y}/{x}.png",
        attribution="© Kartverket",
        html_attribution='&copy; <a href="https://kartverket.no">Kartverket</a>',
    ),
    topogråtone=TileProvider(
        name="Topografisk norgeskart gråtone",
        url="https://cache.kartverket.no/v1/wmts/1.0.0/topograatone/default/webmercator/{z}/{y}/{x}.png",
        attribution="© Kartverket",
        html_attribution='&copy; <a href="https://kartverket.no">Kartverket</a>',
    ),
    toporaster=TileProvider(
        name="Topografisk raster",
        url="https://cache.kartverket.no/v1/wmts/1.0.0/toporaster/default/webmercator/{z}/{y}/{x}.png",
        attribution="© Kartverket",
        html_attribution='&copy; <a href="https://kartverket.no">Kartverket</a>',
    ),
    sjøkart=TileProvider(
        name="Sjøkart",
        url="https://cache.kartverket.no/v1/wmts/1.0.0/sjokartraster/default/webmercator/{z}/{y}/{x}.png",
        attribution="© Kartverket",
        html_attribution='&copy; <a href="https://kartverket.no">Kartverket</a>',
    ),
    norge_i_bilder=TileProvider(
        name="Norge i bilder",
        url="https://opencache.statkart.no/gatekeeper/gk/gk.open_nib_web_mercator_wmts_v2?SERVICE=WMTS&REQUEST=GetTile&VERSION=1.0.0&LAYER=Nibcache_web_mercator_v2&STYLE=default&FORMAT=image/jpgpng&tileMatrixSet=default028mm&tileMatrix={z}&tileRow={y}&tileCol={x}",
        max_zoom=19,
        attribution="© Geovekst",
    ),
)

xyz = Bunch({"Kartverket": kartverket} | providers)



"""Interactive or static map of one or more GeoDataFrames.

This module holds the Map class, which is the basis for the Explore class.
"""

import warnings
from collections.abc import Sequence
from statistics import mean
from typing import Any

import matplotlib
import matplotlib.colors as colors
import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from pandas.api.types import is_dict_like

try:
    from jenkspy import jenks_breaks
except ImportError:
    pass
from pandas.errors import PerformanceWarning
from shapely import Geometry

from ..geopandas_tools.conversion import to_gdf
from ..geopandas_tools.general import _rename_geometry_if
from ..geopandas_tools.general import clean_geoms
from ..geopandas_tools.general import drop_inactive_geometry_columns
from ..geopandas_tools.general import get_common_crs
from ..helpers import get_object_name
from ..helpers import unit_is_meters
from ..raster.image_collection import Band
from ..raster.image_collection import Image
from ..raster.image_collection import ImageCollection

# the geopandas._explore raises a deprication warning. Ignoring for now.
warnings.filterwarnings(
    action="ignore", category=matplotlib.MatplotlibDeprecationWarning
)
warnings.filterwarnings(action="ignore", category=PerformanceWarning)

pd.options.mode.chained_assignment = None


# custom default colors for non-numeric data, because the geopandas default has very
# similar colors. The palette is like the "Set2" cmap from matplotlib, but with more
# colors. If more than 14 categories, the geopandas default cmap is used.
_CATEGORICAL_CMAP = {
    0: "#3b93ff",
    1: "#ff3370",
    2: "#f7cf19",
    3: "#60e825",
    4: "#ff8cc9",
    5: "#804e00",
    6: "#e3dc00",
    7: "#00ffee",
    9: "#870062",
    10: "#751500",
    11: "#1c6b00",
    8: "#7cebb9",
}

DEFAULT_SCHEME = "quantiles"


def proper_fillna(val: Any, fill_val: Any) -> Any:
    """Manually handle missing values when fillna doesn't work as expected.

    Args:
        val: The value to check and fill.
        fill_val: The value to fill in.

    Returns:
        The original value or the filled value if conditions are met.
    """
    try:
        if "NAType" in val.__class__.__name__:
            return fill_val
    except Exception:
        if fill_val is None:
            return fill_val
        if fill_val != fill_val:
            return fill_val

    return val


class Map:
    """Base class that prepares one or more GeoDataFrames for mapping.

    The class has no public methods.
    """

    def __init__(
        self,
        *gdfs: GeoDataFrame,
        column: str | None = None,
        k: int = 5,
        bins: tuple[float] | None = None,
        nan_label: str = "Missing",
        nan_color: str | None = None,
        scheme: str = DEFAULT_SCHEME,
        cmap: str | None = None,
        categorical: bool | None = None,
        **kwargs,
    ) -> None:
        """Initialiser.

        Args:
            *gdfs: Variable length GeoDataFrame list.
            column: The column name to work with.
            k: Number of bins or classes for classification (default: 5).
            bins: Predefined bins for data classification.
            nan_label: Label for missing data.
            nan_color: Color for missing data.
            scheme: Classification scheme to be used.
            cmap (str): Colormap of the plot. See:
                https://matplotlib.org/stable/tutorials/colors/colormaps.html
            categorical: Set to True to convert 'column' to string values.
            **kwargs: Arbitrary keyword arguments.
        """
        self.kwargs = {}

        gdfs, column, kwargs = self._separate_args(gdfs, column, kwargs)

        self._column = column
        self.bins = bins
        self._k = k
        self.nan_label = nan_label
        self.nan_color = nan_color
        self._nan_color_was_none = nan_color is None
        self._cmap = cmap
        self.scheme = scheme
        self._categories_colors_dict = {}
        self._more_data = {}

        # need to get the object names of the gdfs before copying. Only getting,
        # not setting, labels. So the original gdfs don't get the label column.
        self._gdfs: dict[str, GeoDataFrame] = {
            _determine_best_name(gdf, column, i): gdf for i, gdf in enumerate(gdfs)
        }

        show = kwargs.pop("show", True)
        if isinstance(show, (int, bool)):
            self.show = {label: bool(show) for label in self._gdfs}
        elif not hasattr(show, "__iter__"):
            raise ValueError(
                "'show' must be boolean or an iterable of boleans same "
                f"length as gdfs ({len(gdfs)}). Got len {len(show)}"
            )
        else:
            self.show = {label: bool(show) for label in self._gdfs}

        show_args = list(self.show.values())[: len(gdfs)]
        # gdfs that are in kwargs
        show_kwargs = list(self.show.values())[len(gdfs) :]
        for (label, gdf), show in zip(self._gdfs.items(), show_args, strict=False):
            if not len(gdf):
                continue

            gdf = clean_geoms(to_gdf(gdf)).reset_index(drop=True)
            if not len(gdf):
                continue
            self._gdfs[label] = gdf
            self.show[label] = show

        # pop all geometry-like items from kwargs into self._gdfs
        i = 0
        for key, value in kwargs.items():
            try:
                if isinstance(value, Geometry):
                    value = to_gdf(value)
                if not len(value):
                    continue
                self._gdfs[key] = to_gdf(value)
                try:
                    show = show_kwargs[i]
                    i += 1
                except IndexError:
                    pass
                self.show[key] = show
            except Exception:
                self.kwargs[key] = value

        if hasattr(self.show, "__iter__") and len(self.show) != len(self._gdfs):
            raise ValueError(
                "'show' must be boolean or an iterable of boleans same "
                f"length as gdfs ({len(gdfs)}). Got len {len(show)}"
            )

        if categorical is not None:
            self._is_categorical = categorical

        if not self._gdfs or not any(len(gdf) for gdf in self._gdfs.values()):
            self._gdfs = {}
            self._gdf = self._get_gdf_template()
            if categorical is None:
                self._is_categorical = True
            self._unique_values = []
            self._column = "label"
            return

        self._gdfs = self._to_common_crs_and_one_geom_col(self._gdfs)
        if categorical is None:
            self._is_categorical = self._check_if_categorical()

        for label, gdf in self._gdfs.items():
            gdf["label"] = label
            self._gdfs[label] = gdf

        if self._column:
            self._fillna_if_col_is_missing()
        else:
            self._column = "label"

        try:
            self._gdf = pd.concat(self._gdfs.values(), ignore_index=True)
        except ValueError:
            crs = get_common_crs(self._gdfs.values())
            for gdf in self._gdfs.values():
                gdf.crs = crs
            self._gdf = pd.concat(self._gdfs, ignore_index=True)

        self._to_categorical()
        self._get_unique_values()

    def _get_gdf_template(self):
        return GeoDataFrame(
            {
                "geometry": [],
                "label": [],
                "color": [],
            }
            | ({self._column: []} if self._column is not None else {})
        )

    def _prepare_continous_map(self) -> None:
        """Create bins if not already done and adjust k if needed."""
        if self.scheme is None:
            return

        if self.bins is None:
            self.bins = self._create_bins(self._gdf, self._column)
            if len(self.bins) <= self._k and len(self.bins) != len(self._unique_values):
                self._k = len(self.bins)
        elif not all(self._gdf[self._column].isna()):
            self.bins = self._add_minmax_to_bins(self.bins)
            if len(self._unique_values) <= len(self.bins):
                self._k = len(self.bins)  # - 1
        else:
            self._unique_values = self.nan_label
            self._k = 1

    def _prepare_categorical_plot(self):
        """Map values to colors."""
        self._make_categories_colors_dict()
        if self._gdf is not None and len(self._gdf):
            self._fix_nans()

        def _map(value, label):
            try:
                return self._categories_colors_dict[value]
            except KeyError as e:
                if label in self._categories_colors_dict:
                    return self._categories_colors_dict[label]
                if not pd.isna(value):
                    raise e
            return self.nan_color

        for label, gdf in self._gdfs.items():
            gdf["color"] = [_map(value, label) for value in gdf[self._column]]
            self._gdfs[label] = gdf
        self._gdf["color"] = [
            _map(value, label)
            for value, label in zip(
                self._gdf[self._column], self._gdf["label"], strict=False
            )
        ]

    def _to_categorical(self):
        if not (self._is_categorical and self.column is not None):
            return

        def to_string_via_int(series):
            if not pd.api.types.is_numeric_dtype(series):
                return series.astype("string")
            try:
                series = series.astype(float)
            except ValueError:
                return series
            no_decimals: bool = (series.dropna() % 1 == 0).all()
            if no_decimals:
                return series.astype("Int64").astype("string")
            else:
                return series.astype("string")

        for label, gdf in self._gdfs.items():
            if self.column in gdf:
                gdf[self.column] = to_string_via_int(gdf[self.column])
                self._gdfs[label] = gdf
        self._gdf[self.column] = to_string_via_int(self._gdf[self.column])

    def __bool__(self) -> bool:
        """True of any gdfs with more than 0 rows."""
        return bool(len(self._gdfs) + len(self._gdf))

    def _get_unique_values(self) -> None:
        if not self._is_categorical:
            self._unique_values = self._get_unique_floats()
        else:
            unique = list(self._gdf[self._column].unique())
            try:
                self._unique_values = sorted(unique)
            except TypeError:
                self._unique_values = [
                    x for x in unique if proper_fillna(x, None) is not None
                ]

        self._k = min(self._k, len(self._unique_values))

    def _get_unique_floats(self) -> np.array:
        """Get unique floats by multiplying, then converting to integer.

        Find a multiplier that makes the max value greater than +- 1_000_000.
        Because floats don't always equal each other. This will make very
        similar values count as the same value in the color classification.
        """
        array = self._gdf[self._column].dropna()
        self._min = np.min(array)
        self._max = np.max(array)
        self._get_multiplier(array.astype(np.float64))

        unique = array.reset_index(drop=True).drop_duplicates()
        as_int = self._array_to_large_int(unique)
        no_duplicates = as_int.drop_duplicates()

        return np.sort(np.array(unique.loc[no_duplicates.index]))

    def _array_to_large_int(self, array: np.ndarray | pd.Series) -> pd.Series:
        """Multiply values in float array, then convert to integer."""
        if not isinstance(array, pd.Series):
            array = pd.Series(array)

        notna = array[array.notna()]
        isna = array[array.isna()]

        unique_multiplied = (notna.astype(np.float64) * self._multiplier).astype(
            np.int64
        )

        return pd.concat([unique_multiplied, isna]).sort_index()

    def _get_multiplier(self, array: np.ndarray) -> None:
        """Find the number of zeros needed to push the max value of the array above +-1_000_000.

        Adding this as an attribute to use later in _classify_from_bins.
        """
        if not len(array) or np.max(array) == 0:
            self._multiplier: int = 1
            return

        multiplier = 10
        max_ = np.max(array * multiplier)

        if self._max > 0:
            while max_ < 1_000_000:
                multiplier *= 10
                max_ = np.max(array * multiplier)
        else:
            while max_ > -1_000_000:
                multiplier *= 10
                max_ = np.max(array * multiplier)

        self._multiplier: int = multiplier

    def _add_minmax_to_bins(self, bins: list[float | int]) -> list[float | int]:
        """If values are outside the bin range, add max and/or min values of array."""
        # make sure they are lists
        bins = [bin_ for bin_ in bins]

        if min(bins) > 0 and (self._gdf[self._column].dropna().min()) < min(bins):
            num = self._gdf[self._column].dropna().min()
            bins = [num] + bins

        if min(bins) < 0 and (self._gdf[self._column].dropna().min()) < min(bins):
            num = self._gdf[self._column].dropna().min()

            bins = [num] + bins

        if max(bins) > 0 and (self._gdf[self._column].dropna().max()) > max(bins):
            num = self._gdf[self._column].dropna().max()
            bins = bins + [num]

        if max(bins) < 0 and max(
            self._gdf.loc[self._gdf[self._column].notna(), self._column]
        ) < max(bins):
            num = max(self._gdf.loc[self._gdf[self._column].notna(), self._column])
            bins = bins + [num]

        def adjust_bin(num: int | float, i: int) -> int | float:
            if isinstance(num, int):
                return num
            adjuster = (
                float(f"1e-{abs(self.legend.rounding)}")
                if self.legend and self.legend.rounding
                else 0
            )
            if i == 0:
                return num - adjuster
            elif i == len(bins) - 1:
                return num + adjuster
            return num

        bins = [adjust_bin(x, i) for i, x in enumerate(bins)]

        return bins

    @staticmethod
    def _separate_args(
        args: tuple,
        column: str | None,
        kwargs: dict,
    ) -> tuple[tuple[GeoDataFrame], str]:
        """Separate GeoDataFrames from string (column argument)."""

        def as_dict(obj) -> dict:
            if hasattr(obj, "__dict__"):
                return obj.__dict__
            elif isinstance(obj, dict):
                return obj
            raise TypeError(type(obj))

        allowed_types = (
            GeoDataFrame,
            GeoSeries,
            Geometry,
            ImageCollection,
            Image,
            Band,
        )

        gdfs = ()
        more_gdfs = {}
        i = 0
        for arg in args:
            if isinstance(arg, str):
                if column is None:
                    column = arg
                else:
                    raise ValueError(
                        "Can specify at most one string as a positional argument."
                    )
            elif isinstance(arg, allowed_types):
                gdfs = gdfs + (arg,)
            # elif isinstance(arg, Sequence) and not isinstance(arg, str):
            elif isinstance(arg, dict) or hasattr(arg, "__dict__"):
                # add dicts or classes with GeoDataFrames to kwargs
                for key, value in as_dict(arg).items():
                    if isinstance(value, allowed_types):
                        more_gdfs[key] = value
                    elif isinstance(value, dict) or hasattr(value, "__dict__"):
                        # elif isinstance(value, Sequence) and not isinstance(value, str):
                        try:
                            # same as above, one level down
                            more_gdfs |= {
                                k: v
                                for k, v in as_dict(value).items()
                                if isinstance(v, allowed_types)
                            }
                        except Exception:
                            # ignore all exceptions
                            pass

            elif isinstance(arg, Sequence) and not isinstance(arg, str):
                # add dicts or classes with GeoDataFrames to kwargs
                for value in arg:
                    if isinstance(value, allowed_types):
                        name = _determine_best_name(value, column, i)
                        more_gdfs[name] = value
                    elif isinstance(value, dict) or hasattr(value, "__dict__"):
                        try:
                            # same as above, one level down
                            more_gdfs |= {
                                k: v
                                for k, v in value.items()
                                if isinstance(v, allowed_types)
                            }
                        except Exception:
                            # no need to raise here
                            pass
                    elif isinstance(value, Sequence) and not isinstance(value, str):
                        for x in value:
                            if not isinstance(x, allowed_types):
                                continue
                            name = _determine_best_name(value, column, i)
                            more_gdfs[name] = x
                    i += 1

        kwargs |= more_gdfs

        return gdfs, column, kwargs

    def _to_common_crs_and_one_geom_col(
        self, gdfs: dict[str, GeoDataFrame]
    ) -> dict[str, GeoDataFrame]:
        """Need common crs and max one geometry column."""
        crs_list = list({gdf.crs for gdf in gdfs.values() if gdf.crs is not None})
        if crs_list:
            self.crs = crs_list[0]
        for label, gdf in gdfs.items():
            gdf = gdf.reset_index(drop=True)
            gdf = drop_inactive_geometry_columns(gdf).pipe(_rename_geometry_if)
            if crs_list:
                try:
                    gdf = gdf.to_crs(self.crs)
                except ValueError:
                    gdf = gdf.set_crs(self.crs)
            gdfs[label] = gdf
        return gdfs

    def _fillna_if_col_is_missing(self) -> None:
        n = 0
        for label, gdf in self._gdfs.items():
            if self._column in gdf.columns:
                gdf[self._column] = gdf[self._column].fillna(pd.NA)
                n += 1
            else:
                gdf[self._column] = pd.NA
            self._gdfs[label] = gdf

        maybe_area = 1 if "area" in self._column else 0
        maybe_length = (
            1 if any(x in self._column for x in ["meter", "metre", "leng"]) else 0
        )
        n = n + maybe_area + maybe_length

    def _check_if_categorical(self) -> bool:
        """Quite messy this..."""
        if not self._column or not self._gdfs:
            return True

        def is_maybe_km2():
            if "area" in self._column and (
                "km2" in self._column
                or (
                    "kilomet" in self._column
                    and ("sq" in self._column or "2" in self._column)
                )
            ):
                return True
            else:
                return False

        maybe_area = 1 if "area" in self._column else 0
        maybe_area_km2 = 1 if is_maybe_km2() else 0
        maybe_length = (
            1 if any(x in self._column for x in ["meter", "metre", "leng"]) else 0
        )

        all_nan = 0
        col_not_present = 0
        for gdf in self._gdfs.values():
            if self._column not in gdf:
                if maybe_area_km2 and unit_is_meters(gdf):
                    gdf["area_km2"] = gdf.area / 1_000_000
                    maybe_area_km2 += 1
                elif maybe_area:
                    gdf["area"] = gdf.area
                    maybe_area += 1
                elif maybe_length:
                    gdf["length"] = gdf.length
                    maybe_length += 1
                else:
                    col_not_present += 1
            elif not pd.api.types.is_numeric_dtype(gdf[self._column]):
                if all(gdf[self._column].isna()):
                    all_nan += 1
                return True

        if maybe_area_km2 > 1:
            self._column = "area_km2"
            return False
        if maybe_area > 1:
            self._column = "area"
            return False
        if maybe_length > 1:
            self._column = "length"
            return False

        if all_nan == len(self._gdfs):
            raise ValueError(
                f"All values are NaN in column {self.column!r}. {self._gdfs}"
            )

        if col_not_present == len(self._gdfs):
            raise ValueError(f"{self.column} not found.")

        return False

    def _make_categories_colors_dict(self) -> None:
        if "color" in self.kwargs and is_dict_like(self.kwargs["color"]):
            if self._column is None and not all(
                key in self.kwargs["color"] for key in self._gdfs
            ):
                raise ValueError(
                    "When specifying 'color' as dict-like, you must also pass a column "
                    "or all gdfs passed must have labels/names corresponding to keys in the color dict."
                )
            self._categories_colors_dict = self.kwargs.pop("color")
        elif not self._cmap and len(self._unique_values) <= len(_CATEGORICAL_CMAP):
            # custom categorical cmap
            self._categories_colors_dict = {
                category: _CATEGORICAL_CMAP[i]
                for i, category in enumerate(self._unique_values)
            } | self._categories_colors_dict
        elif self._cmap:
            cmap = matplotlib.colormaps.get_cmap(self._cmap)
            self._categories_colors_dict = {
                category: colors.to_hex(cmap(int(i)))
                for i, category in enumerate(self._unique_values)
            } | self._categories_colors_dict
        else:
            cmap = matplotlib.colormaps.get_cmap("tab20")
            self._categories_colors_dict = {
                category: colors.to_hex(cmap(int(i)))
                for i, category in enumerate(self._unique_values)
            } | self._categories_colors_dict

    def _fix_nans(self) -> None:
        def hasnans(df: GeoDataFrame, label: str) -> bool:
            if label in self._categories_colors_dict:
                return False
            return df[self._column].isna().any()

        if any(hasnans(df, label) for label, df in self._gdfs.items()):
            self._gdf[self._column] = self._gdf[self._column].fillna(self.nan_label)
            self._categories_colors_dict[self.nan_label] = (
                self._categories_colors_dict.get(self.nan_label, self.nan_color)
            )

    def _create_bins(self, gdf: GeoDataFrame, column: str) -> np.ndarray:
        """Make bin list of length k + 1, or length of unique values.

        The returned bins sometimes have two almost identical

        If 'scheme' is not specified, the jenks_breaks function is used, which is
        much faster than the one from Mapclassifier.
        """
        if not gdf[column].notna().any():
            return np.array([0])

        n_classes = (
            self._k if len(self._unique_values) > self._k else len(self._unique_values)
        )

        if self._k == len(self._unique_values) - 1:
            n_classes = self._k - 1
            self._k = self._k - 1

        if self._k > len(self._unique_values):
            self._k = len(self._unique_values)
            n_classes = len(self._unique_values)

        if self.scheme == "jenks":
            bins = jenks_breaks(gdf[column].dropna(), n_classes=n_classes)
        else:
            # local import because slow
            from mapclassify import classify

            binning = classify(
                np.asarray(gdf[column].dropna()),
                scheme=self.scheme,
                # k=self._k,
                k=n_classes,
            )
            bins = binning.bins

        bins = self._add_minmax_to_bins(bins)

        unique_bins = list({round(bin_, 5) for bin_ in bins})
        unique_bins.sort()

        if self._k == len(self._unique_values) - 1 or len(unique_bins) == len(
            self._unique_values
        ):
            return np.array(unique_bins)

        if len(unique_bins) == len(bins) - 1:
            self._k -= 1

        return np.array(bins)

    def change_cmap(self, cmap: str, start: int = 0, stop: int = 256) -> "Map":
        """Change the color palette of the plot.

        Args:
            cmap: The colormap.
                https://matplotlib.org/stable/tutorials/colors/colormaps.html
            start: Start position for the color palette. Defaults to 0.
            stop: End position for the color palette. Defaults to 256, which
                is the end of the color range.
        """
        self.cmap_start = start
        self.cmap_stop = stop
        self._cmap = cmap
        self._cmap_has_been_set = True
        return self

    def _get_continous_colors(self, n: int) -> np.ndarray:
        cmap = matplotlib.colormaps.get_cmap(self._cmap)
        colors_ = [
            colors.to_hex(cmap(int(i)))
            for i in np.linspace(self.cmap_start, self.cmap_stop, num=n)
        ]
        if self._gdf[self._column].isna().any():
            colors_ = colors_ + [self.nan_color]
        return np.array(colors_)

    def _classify_from_bins(self, gdf: GeoDataFrame, bins: np.ndarray) -> np.ndarray:
        """Place the column values into groups."""
        bins = bins.copy()

        assert gdf.index.is_unique

        # if equal length, convert to integer and check for equality
        if len(bins) == len(self._unique_values):
            if gdf[self._column].isna().all():
                return np.repeat(len(bins), len(gdf))

            gdf["col_as_int"] = self._array_to_large_int(gdf[self._column])
            bins = self._array_to_large_int(self._unique_values)
            gdf["col_as_int"] = gdf["col_as_int"].fillna(np.nan)
            classified = np.searchsorted(bins, gdf["col_as_int"])

        else:
            if len(bins) == self._k + 1:
                bins = bins[1:]

            if (
                self.legend
                and self.legend.rounding
                and (self.legend.rounding or 1) <= 0
            ):
                bins[0] = bins[0] - 1
                bins[-1] = bins[-1] + 1

            if gdf[self._column].isna().all():
                return np.repeat(len(bins), len(gdf))

            # need numpy.nan instead of pd.NA
            gdf[self._column] = gdf[self._column].fillna(np.nan)

            gdf[self._column] = gdf[self._column].apply(
                lambda x: proper_fillna(x, np.nan)
            )

            classified = np.searchsorted(bins, gdf[self._column])

        return classified

    def _push_classification(self, classified: np.ndarray) -> np.ndarray:
        """Push classes downwards if gaps in classification sequence.

        So from e.g. [0,2,4] to [0,1,2].

        Otherwise, we will get index error when classifying colors.
        """
        rank_dict = {val: rank for rank, val in enumerate(np.unique(classified))}

        return np.array([rank_dict[val] for val in classified])

    @property
    def k(self) -> int:
        """Number of bins."""
        return self._k

    @k.setter
    def k(self, new_value: int) -> None:
        if not self._is_categorical and new_value > len(self._unique_values):
            raise ValueError(
                "'k' cannot be greater than the number of unique values in the column.'"
                f"Got new k={new_value} and previous k={len(self._unique_values)}.'"
                #  f"{''.join(self._unique_values)}"
            )
        self._k = int(new_value)

    @property
    def cmap(self) -> str:
        """Colormap."""
        return self._cmap

    @cmap.setter
    def cmap(self, new_value: str) -> None:
        self._cmap = new_value
        if not self._is_categorical:
            self.change_cmap(cmap=new_value, start=self.cmap_start, stop=self.cmap_stop)

    @property
    def gdf(self) -> GeoDataFrame:
        """All GeoDataFrames concated."""
        return self._gdf

    @gdf.setter
    def gdf(self, _) -> None:
        raise ValueError(
            "Cannot change 'gdf' after init. Put the GeoDataFrames into "
            "the class initialiser."
        )

    @property
    def gdfs(self) -> list[GeoDataFrame]:
        """All GeoDataFrames as a list."""
        return self._gdfs

    @gdfs.setter
    def gdfs(self, _) -> None:
        raise ValueError(
            "Cannot change 'gdfs' after init. Put the GeoDataFrames into "
            "the class initialiser."
        )

    @property
    def column(self) -> str | None:
        """Column to use as colormap."""
        return self._column

    @column.setter
    def column(self, _) -> None:
        raise ValueError(
            "Cannot change 'column' after init. Specify 'column' in the "
            "class initialiser."
        )

    def __setitem__(self, item: Any, new_item: Any) -> None:
        """Set an attribute with square brackets."""
        return setattr(self, item, new_item)

    def __getitem__(self, item: Any) -> Any:
        """Get an attribute with square brackets."""
        return getattr(self, item)

    def get(self, key: Any, default: Any | None = None) -> Any:
        """Get an attribute with default value if not present."""
        try:
            return self[key]
        except (KeyError, ValueError, IndexError, AttributeError):
            return default


def _determine_best_name(obj: Any, column: str | None, i: int) -> str:
    try:
        # Frame 3: actual object name Frame 2: maps.py:explore(). Frame 1: __init__. Frame 0: this function.
        return str(get_object_name(obj, start=3))
    except ValueError:
        if isinstance(obj, GeoSeries) and obj.name:
            return str(obj.name)
        elif isinstance(obj, GeoDataFrame) and len(obj.columns) == 2 and not column:
            series = obj.drop(columns=obj._geometry_column_name).iloc[:, 0]
            if (
                len(series.unique()) == 1
                and mean(isinstance(x, str) for x in series) > 0.5
            ):
                return str(next(iter(series)))
            elif series.name:
                return str(series.name)
        else:
            # generic label e.g. Image(1)
            return f"{obj.__class__.__name__}({i})"



"""Cutting and splitting line geometries."""

import warnings

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from pandas import DataFrame
from pandas import Series
from shapely import force_2d
from shapely.geometry import LineString
from shapely.geometry import Point

from ..geopandas_tools.general import _split_lines_by_points_along_line
from ..geopandas_tools.geometry_types import get_geom_type
from ..geopandas_tools.point_operations import snap_all
from ..geopandas_tools.point_operations import snap_within_distance

PRECISION = 1e-6


def split_lines_by_nearest_point(
    gdf: GeoDataFrame,
    points: GeoDataFrame,
    max_distance: int | float | None = None,
    splitted_col: str | None = None,
) -> DataFrame:
    """Split lines that are closest to s point.

    Snaps points to nearest lines and splits the lines in two at the snap point.
    The splitting is done pointwise, meaning each point splits one line in two.
    The line will not be split if the point is closest to the endpoint of the line.

    This function is used in NetworkAnalysis if split_lines is set to True.

    Args:
        gdf: GeoDataFrame of lines that will be split.
        points: GeoDataFrame of points to split the lines with.
        max_distance: the maximum distance between the point and the line. Points
            further away than max_distance will not split any lines. Defaults to None.
        splitted_col: Optionally add a column

    Returns:
        A GeoDataFrame with the same columns as the input lines, but with the lines
        split at the closest point to the points. If no lines are within 'max_distance'
        from any points, 'gdf' is returned as it came.

    Raises:
        ValueError: If the crs of the input data differs.

    Examples:
    ---------
    >>> from sgis import read_parquet_url, split_lines_by_nearest_point
    >>> roads = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")
    >>> points = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/points_oslo.parquet")
    >>> rows = len(roads)
    >>> rows
    93395

    Splitting lines for points closer than 10 meters from the lines.

    >>> roads = split_lines_by_nearest_point(roads, points, max_distance=10)
    >>> print("number of lines that were split:", len(roads) - rows)
    number of lines that were split: 380

    Splitting lines by all points.

    >>> roads = split_lines_by_nearest_point(roads, points)
    >>> print("number of lines that were split:", len(roads) - rows)
    number of lines that were split: 848

    Not all lines were split. That is because some points were closest to an endpoint
    of a line.
    """
    if not len(gdf) or not len(points):
        return gdf

    if (points.crs is not None and gdf.crs is not None) and not points.crs.equals(
        gdf.crs
    ):
        raise ValueError("crs mismatch:", points.crs, "and", gdf.crs)

    if get_geom_type(gdf) != "line":
        raise ValueError(
            f"'gdf' should only have line geometriess. Got {gdf.geom_type.value_counts()}"
        )

    if get_geom_type(points) != "point":
        raise ValueError(
            f"'points' should only have point geometries. Got {points.geom_type.value_counts()}"
        )

    gdf = gdf.copy()

    # move the points to the nearest exact location on the line
    if max_distance:
        snapped = snap_within_distance(points, gdf, max_distance=max_distance)
    else:
        snapped = snap_all(points, gdf)

    return _split_lines_by_points_along_line(gdf, snapped, splitted_col=splitted_col)


def cut_lines(
    gdf: GeoDataFrame, max_length: int, ignore_index: bool = False
) -> GeoDataFrame:
    """Cuts lines of a GeoDataFrame into pieces of a given length.

    Args:
        gdf: GeoDataFrame.
        max_length: The maximum length of the lines in the output GeoDataFrame.
        ignore_index: If True, the resulting axis will be labeled 0, 1, …, n - 1.
            Defaults to False.

    Returns:
        A GeoDataFrame with lines cut to the maximum distance.

    Note:
        This method is time consuming for large networks and low 'max_length'.

    Examples:
    ---------
    >>> from sgis import read_parquet_url, cut_lines
    >>> roads = read_parquet_url("https://media.githubusercontent.com/media/statisticsnorway/ssb-sgis/main/tests/testdata/roads_oslo_2022.parquet")
    >>> roads.length.describe().round(1)
    count    93395.0
    mean        41.2
    std         78.5
    min          0.2
    25%         14.0
    50%         27.7
    75%         47.5
    max       5213.7
    dtype: float64
    >>> roads = cut_lines(roads, max_length=100)
    >>> roads.length.describe().round(1)
    count    126304.0
    mean         30.5
    std          30.1
    min           0.0
    25%           5.7
    50%          22.5
    75%          44.7
    max         100.0
    dtype: float64
    """
    if get_geom_type(gdf) != "line":
        raise ValueError("'gdf' should only have line geometries.")

    gdf["geometry"] = force_2d(gdf.geometry)

    gdf = gdf.explode(ignore_index=ignore_index, index_parts=False)

    long_lines = gdf.loc[gdf.length > max_length]

    if not len(long_lines):
        return gdf

    for x in [10, 5, 1]:
        max_ = max(long_lines.length)
        while max_ > max_length * x + 1:
            max_ = max(long_lines.length)

            long_lines = cut_lines_once(long_lines, max_length)

            if max_ == max(long_lines.length):
                break

    long_lines = long_lines.explode(ignore_index=ignore_index, index_parts=False)

    short_lines = gdf.loc[gdf.length <= max_length]

    return pd.concat([short_lines, long_lines], ignore_index=ignore_index)


def cut_lines_once(
    gdf: GeoDataFrame,
    distances: int | float | str | Series,
    ignore_index: bool = False,
) -> GeoDataFrame:
    """Cuts lines of a GeoDataFrame in two at the given distance or distances.

    Takes a GeoDataFrame of lines and cuts each line in two. If distances is a number,
    all lines will be cut at the same length.

    Args:
        gdf: GeoDataFrame.
        distances: The distance from the start of the lines to cut at. Either a number,
            the name of a column or array-like of same length as the line GeoDataFrame.
        ignore_index: If True, the resulting axis will be labeled 0, 1, …, n - 1.
            Defaults to False.

    Examples:
    ---------
    >>> from sgis import cut_lines_once, to_gdf
    >>> import pandas as pd
    >>> from shapely.geometry import LineString
    >>> gdf = to_gdf(LineString([(0, 0), (1, 1), (2, 2)]))
    >>> gdf = pd.concat([gdf, gdf, gdf])
    >>> gdf
                                                geometry
    0  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...
    0  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...
    0  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...

    Cut all lines at 1 unit from the start of the lines.

    >>> cut_lines_once(gdf, 1)
                                                geometry
    0      LINESTRING (0.00000 0.00000, 0.70711 0.70711)
    1  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...
    2      LINESTRING (0.00000 0.00000, 0.70711 0.70711)
    3  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...
    4      LINESTRING (0.00000 0.00000, 0.70711 0.70711)
    5  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...

    Cut distance as column.

    >>> gdf["dist"] = [1, 2, 3]
    >>> cut_lines_once(gdf, "dist")
                                                geometry  dist
    0      LINESTRING (0.00000 0.00000, 0.70711 0.70711)     1
    0  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...     1
    0  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...     2
    0      LINESTRING (1.41421 1.41421, 2.00000 2.00000)     2
    0  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...     3

    Cut distance as list (same result as above).

    >>> cut_lines_once(gdf, [1, 2, 3])
                                                geometry  dist
    0      LINESTRING (0.00000 0.00000, 0.70711 0.70711)     1
    0  LINESTRING (0.70711 0.70711, 1.00000 1.00000, ...     1
    0  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...     2
    0      LINESTRING (1.41421 1.41421, 2.00000 2.00000)     2
    0  LINESTRING (0.00000 0.00000, 1.00000 1.00000, ...     3
    """

    def _cut(line: LineString, distance: int | float) -> list[LineString]:
        """From the shapely docs."""
        if distance <= 0.0 or distance >= line.length:
            return line
        coords = list(line.coords)
        for i, p in enumerate(coords):
            prd = line.project(Point(p))
            if prd == distance:
                return [LineString(coords[: i + 1]), LineString(coords[i:])]
            if prd > distance:
                cp = line.interpolate(distance)
                return [
                    LineString(coords[:i] + [(cp.x, cp.y)]),
                    LineString([(cp.x, cp.y)] + coords[i:]),
                ]

    crs = gdf.crs
    geom_col = gdf._geometry_column_name

    gdf = gdf.copy()

    # cutting lines will give lists of linestrings in the geometry column. Ignoring
    # the warning it triggers
    warnings.filterwarnings(
        "ignore", message="Geometry column does not contain geometry."
    )

    if isinstance(distances, str):
        gdf[geom_col] = np.vectorize(_cut)(gdf[geom_col], gdf[distances])
    else:
        gdf[geom_col] = np.vectorize(_cut)(gdf[geom_col], distances)

    # explode will give pandas df if not gdf is constructed
    gdf = GeoDataFrame(
        gdf.explode(ignore_index=ignore_index, index_parts=False),
        geometry=geom_col,
        crs=crs,
    )

    return gdf



import abc
import datetime
import json
import re
from collections.abc import Iterable
from io import BytesIO
from pathlib import Path
from typing import Any
from urllib.request import urlopen

import folium
import numpy as np
import pandas as pd
import shapely
from geopandas import GeoDataFrame
from geopandas import GeoSeries
from shapely import Geometry
from shapely import get_exterior_ring
from shapely import make_valid
from shapely import polygons
from shapely import set_precision
from shapely import simplify
from shapely.errors import GEOSException

from ..geopandas_tools.conversion import to_gdf
from ..geopandas_tools.conversion import to_shapely
from ..geopandas_tools.sfilter import sfilter
from ..raster.image_collection import Band

JSON_PATH = Path(__file__).parent / "norge_i_bilder.json"

# JSON_YEARS = tuple(range(1900, datetime.datetime.now().year + 1))
JSON_YEARS = tuple(range(2006, datetime.datetime.now().year + 1))

DEFAULT_YEARS: tuple[int] = tuple(
    range(
        int(datetime.datetime.now().year) - 10,
        int(datetime.datetime.now().year) + 1,
    )
)


class WmsLoader(abc.ABC):
    """Abstract base class for wms loaders.

    Child classes must implement the method 'get_tiles',
    which should return a list of folium.WmsTileLayer.
    """

    _min_year: int = 1900

    @abc.abstractmethod
    def filter_tiles(
        self, mask: GeoDataFrame | GeoSeries | Geometry | tuple[float]
    ) -> list[str]:
        """Filter relevant dates with pandas and geopandas because fast."""

    @abc.abstractmethod
    def get_tiles(self, bbox: Any, max_zoom: int = 40) -> list[folium.WmsTileLayer]:
        """Get all tiles intersecting with a bbox."""

    @abc.abstractmethod
    def load_tiles(self) -> None:
        """Load all tiles into self._tiles.

        Not needed in sgis.explore.
        """
        pass

    def __repr__(self) -> str:
        """Print representation."""
        return str(self)

    def __str__(self) -> str:
        """String representation."""

        def maybe_to_string(value: Any):
            if isinstance(value, str):
                return f"'{value}'"
            return value

        txt = ", ".join(
            f"{k}={maybe_to_string(v)}"
            for k, v in self.__dict__.items()
            if not k.startswith("_")
        )
        return f"{self.__class__.__name__}({txt})"


class NorgeIBilderWms(WmsLoader):
    """Loads Norge i bilder tiles as folium.WmsTiles.

    Args:
        years: list of years to search for images.
        contains: substrings to include in image search.
        not_contains: substrings to exclude in image search.
        show: Whether to show all layers upon initialisation of the map.
        _use_json: Whether to use pre-made json file of image names and bounds/masks
            if all years are within range. Defaults to True (much faster).
    """

    url: str = "https://wms.geonorge.no/skwms1/wms.nib-prosjekter"

    def __init__(
        self,
        years: Iterable[int | str] = DEFAULT_YEARS,
        contains: str | Iterable[str] | None = None,
        not_contains: str | Iterable[str] | None = None,
        show: bool | Iterable[int] | int = False,
        _use_json: bool = True,
    ) -> None:
        """Initialiser."""
        self.years = [int(year) for year in years]
        self.contains = contains
        self.not_contains = not_contains
        self.show = show
        self._use_json = _use_json

        if self._use_json and all(year in JSON_YEARS for year in self.years):
            self._load_from_json()
        else:
            self._tiles = None

    def load_tiles(self, verbose: bool = False) -> None:
        """Load all Norge i bilder tiles into self._tiles."""
        name_pattern = r"<Name>(.*?)</Name>"
        bbox_pattern = (
            r"<EX_GeographicBoundingBox>.*?"
            r"<westBoundLongitude>(.*?)</westBoundLongitude>.*?"
            r"<eastBoundLongitude>(.*?)</eastBoundLongitude>.*?"
            r"<southBoundLatitude>(.*?)</southBoundLatitude>.*?"
            r"<northBoundLatitude>(.*?)</northBoundLatitude>.*?</EX_GeographicBoundingBox>"
        )

        all_tiles: list[dict] = []
        with urlopen(self.url) as file:
            xml_data: str = file.read().decode("utf-8")

            for text in xml_data.split('<Layer queryable="1">')[1:]:

                # Extract bounding box values
                bbox_match = re.search(bbox_pattern, text, re.DOTALL)
                if bbox_match:
                    minx, maxx, miny, maxy = (
                        float(bbox_match.group(i)) for i in [1, 2, 3, 4]
                    )
                    this_bbox = shapely.box(minx, miny, maxx, maxy)
                else:
                    this_bbox = None

                name_match = re.search(name_pattern, text, re.DOTALL)
                name = name_match.group(1) if name_match else None

                if (
                    not name
                    or not any(str(year) in name for year in self.years)
                    or (
                        self.contains
                        and not any(
                            re.search(x, name.lower())
                            for x in _string_as_list(self.contains)
                        )
                    )
                    or (
                        self.not_contains
                        and any(
                            re.search(x, name.lower())
                            for x in _string_as_list(self.not_contains)
                        )
                    )
                ):
                    continue

                this_tile = {}
                this_tile["name"] = name
                this_tile["bbox"] = this_bbox
                year = name.split(" ")[-1]
                is_year_or_interval: bool = all(
                    part.isnumeric() and len(part) == 4 for part in year.split("-")
                )
                if is_year_or_interval:
                    this_tile["year"] = year
                else:
                    this_tile["year"] = "9999"

                all_tiles.append(this_tile)

        self._tiles = sorted(all_tiles, key=lambda x: (x["year"]))

        masks = self._get_norge_i_bilder_polygon_masks(verbose=verbose)
        for tile in self._tiles:
            mask = masks.get(tile["name"], None)
            tile["geometry"] = mask

    def _get_norge_i_bilder_polygon_masks(self, verbose: bool):
        from owslib.util import ServiceException
        from owslib.wms import WebMapService
        from PIL import Image

        relevant_names: dict[str, str] = {x["name"]: x["bbox"] for x in self._tiles}
        assert len(relevant_names), relevant_names

        url = "https://wms.geonorge.no/skwms1/wms.nib-mosaikk?SERVICE=WMS&REQUEST=GetCapabilities"
        wms = WebMapService(url, version="1.3.0")
        out = {}

        for layer in list(wms.contents):
            title = wms[layer].title
            if title not in relevant_names:
                continue
            bbox = wms[layer].boundingBoxWGS84
            bbox = tuple(to_gdf(bbox, crs=4326).to_crs(25832).total_bounds)

            existing_bbox = relevant_names[title]
            existing_bbox = to_gdf(existing_bbox, crs=4326).to_crs(25832).union_all()
            if not to_shapely(bbox).intersects(existing_bbox):
                continue
            diffx = bbox[2] - bbox[0]
            diffy = bbox[3] - bbox[1]
            width = int(diffx / 40)
            height = int(diffy / 40)
            if not bbox:
                continue
            try:
                img = wms.getmap(
                    layers=[layer],
                    styles=[""],  # Empty unless you know the style
                    srs="EPSG:25832",
                    bbox=bbox,
                    size=(width, height),
                    format="image/jpeg",
                    transparent=True,
                    bgcolor="#FFFFFF",
                )
            except (ServiceException, AttributeError) as e:
                if verbose:
                    print(type(e), e)
                continue

            arr = np.array(Image.open(BytesIO(img.read())))
            if not np.sum(arr):
                continue

            band = Band(
                np.where(np.any(arr != 0, axis=-1), 1, 0), bounds=bbox, crs=25832
            )
            polygon = band.to_geopandas()[lambda x: x["value"] == 1].geometry.values
            polygon = make_valid(polygons(get_exterior_ring(polygon)))
            polygon = make_valid(set_precision(polygon, 1))
            polygon = make_valid(simplify(polygon, 100))
            polygon = make_valid(set_precision(polygon, 1))
            polygon = GeoSeries(polygon, crs=25832).to_crs(4326)
            if verbose:
                print(f"Layer name: {layer}")
                print(f"Title: {wms[layer].title}")
                print(f"Bounding box: {wms[layer].boundingBoxWGS84}")
                print(f"polygon: {polygon}")
                print("-" * 40)

            for x in [0, 0.1, 0.001, 1]:
                try:
                    out[title] = make_valid(polygon.buffer(x).make_valid().union_all())
                except GEOSException:
                    pass
                break

        return out

    def get_tiles(self, mask: Any, max_zoom: int = 40) -> list[folium.WmsTileLayer]:
        """Get all Norge i bilder tiles intersecting with a mask (bbox or polygon)."""
        if self._tiles is None:
            self.load_tiles()

        if not isinstance(mask, (GeoSeries | GeoDataFrame | Geometry)):
            mask = to_shapely(mask)

        if isinstance(self.show, bool):
            show = self.show
        else:
            show = False

        relevant_tiles = self.filter_tiles(mask)
        tile_layers = {
            name: folium.WmsTileLayer(
                url="https://wms.geonorge.no/skwms1/wms.nib-prosjekter",
                name=name,
                layers=name,
                format="image/png",  # Tile format
                transparent=True,  # Allow transparency
                version="1.3.0",  # WMS version
                attr="&copy; <a href='https://www.geonorge.no/'>Geonorge</a>",
                show=show,
                max_zoom=max_zoom,
            )
            for name in relevant_tiles
        }

        if not len(tile_layers):
            return tile_layers

        if isinstance(self.show, int) and not isinstance(self.show, bool):
            tile = tile_layers[list(tile_layers)[self.show]]
            tile.show = True
        elif isinstance(self.show, Iterable):
            for i in self.show:
                tile = tile_layers[list(tile_layers)[i]]
                tile.show = True

        return tile_layers

    def filter_tiles(
        self, mask: GeoDataFrame | GeoSeries | Geometry | tuple[float]
    ) -> list[str]:
        """Filter relevant dates with pandas and geopandas because fast."""
        if not self._tiles:
            return []
        df = pd.DataFrame(self._tiles)
        filt = (df["name"].notna()) & (
            df["year"].str.contains("|".join([str(year) for year in self.years]))
        )
        if self.contains:
            for x in _string_as_list(self.contains):
                filt &= df["name"].str.contains(x)
        if self.not_contains:
            for x in _string_as_list(self.not_contains):
                filt &= ~df["name"].str.contains(x)
        df = df[filt]
        geoms = np.where(df["geometry"].notna(), df["geometry"], df["bbox"])
        geoms = GeoSeries(geoms)
        assert geoms.index.is_unique
        return list(df.iloc[sfilter(geoms, mask).index]["name"])

    def _load_from_json(self) -> None:
        """Load tiles from json file."""
        try:
            with open(JSON_PATH, encoding="utf-8") as file:
                self._tiles = json.load(file)
        except FileNotFoundError:
            self._tiles = None
            return
        self._tiles = [
            {
                key: (
                    value
                    if key not in ["bbox", "geometry"]
                    else shapely.wkt.loads(value)
                )
                for key, value in tile.items()
            }
            for tile in self._tiles
            if any(str(year) in tile["name"] for year in self.years)
        ]


def _string_as_list(x: str | list[str]) -> list[str] | None:
    if not x:
        return None
    return [x] if isinstance(x, str) else list(x)
